(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 6.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[    519198,      10052]
NotebookOptionsPosition[    513671,       9867]
NotebookOutlinePosition[    514108,       9884]
CellTagsIndexPosition[    514065,       9881]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{

Cell[CellGroupData[{
Cell["Test of Model Predictions", "Title"],

Cell["\<\
This Notebook shows how REAP can be used to test how the predictions of a \
given model change during the RG evolution to low energy.\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"If", "[", 
   RowBox[{
    RowBox[{"$VersionNumber", "<", "6"}], ",", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Get", "[", "\"\<Graphics`FilledPlot`\>\"", "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"Get", "[", "\"\<Graphics`Colors`\>\"", "]"}]}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"ForestGreen", "=", 
      RowBox[{"RGBColor", "[", 
       RowBox[{"0.1", ",", "0.75", ",", "0.2"}], "]"}]}], ";"}]}], 
   "\[IndentingNewLine]", "]"}], ";"}]], "Input",
 CellChangeTimes->{{3.435587848557613*^9, 3.435587856517888*^9}, {
  3.435655531948548*^9, 3.435655544087432*^9}, {3.635817659548285*^9, 
  3.6358176642028017`*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Needs", "[", "\"\<REAP`RGEMSSM`\>\"", "]"}], ";"}], "\n", 
 RowBox[{
  RowBox[{"Needs", "[", "\"\<REAP`RGEPlotUtilities`\>\"", "]"}], 
  ";"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData["\<\"Loading REAP 1.11\\n\"\>"], "Print",
 CellChangeTimes->{3.646477798163739*^9}],

Cell[BoxData["\<\"SM0N: Two loop RG equation of \[Kappa] not implemented yet.\
\"\>"], "Print",
 CellChangeTimes->{3.646477798215672*^9}],

Cell[BoxData["\<\"SM: Two loop RG equations of Y\[Nu], M\[Nu]r and \[Kappa] \
not implemented yet.\"\>"], "Print",
 CellChangeTimes->{3.6464777982202377`*^9}]
}, Open  ]]
}, Open  ]],

Cell[BoxData[{
 RowBox[{
  RowBox[{"MZ", "=", "91.12"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"GUT", "=", 
   RowBox[{"2", " ", 
    RowBox[{"10", "^", "16"}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.6368469804624157`*^9, 3.63684698357728*^9}}],

Cell["\<\
RGEReset is used in order to set all variables of RGESolver to their default \
values, if the notebook is executed several times.\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RGEReset", "[", "]"}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell["Definition of the model", "Section"],

Cell["\<\
We assume a supersymmetric model.  Below the SUSY-breaking scale (taken to be \
1 TeV), the SM is used as an effective theory.\
\>", "Text"],

Cell[BoxData[
 RowBox[{"RGEReset", "[", "]"}]], "Input",
 CellChangeTimes->{{3.635818682949304*^9, 3.63581868546563*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"RGEAdd", "[", 
   RowBox[{"\"\<MSSM\>\"", ",", 
    RowBox[{"RGEtan\[Beta]", "\[Rule]", "50"}], ",", 
    RowBox[{"RGESearchTransitions", "\[Rule]", "True"}]}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"RGEAdd", "[", 
   RowBox[{"\"\<SM\>\"", ",", 
    RowBox[{"RGECutoff", "\[Rule]", "1000"}]}], "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.6358186882106533`*^9, 3.635818699794078*^9}}],

Cell[TextData[{
 "All couplings are given explicitly and no default values of the REAP \
package are used. The Yukawa couplings are given in \"RL convention\".  The \
neutrino masses are determined by ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["Y", "\[Nu]"], TraditionalForm]]],
 "and ",
 Cell[BoxData[
  FormBox["M", TraditionalForm]]],
 ".\n\[Kappa] is set to 0. In a type II see-saw scenario, \[Kappa] is \
different from 0 below the mass threshold of the Higgs triplet."
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RGESetInitial", "[", 
   RowBox[{"GUT", ",", "\[IndentingNewLine]", 
    RowBox[{"RGEg1", "\[Rule]", "0.7044"}], ",", "\[IndentingNewLine]", 
    RowBox[{"RGEg2", "\[Rule]", "0.6965"}], ",", "\[IndentingNewLine]", 
    RowBox[{"RGEg3", "\[Rule]", "0.6984"}], ",", "\[IndentingNewLine]", 
    RowBox[{"RGEYu", "\[Rule]", 
     RowBox[{"(", "\[NoBreak]", GridBox[{
        {
         RowBox[{"5.980", " ", 
          SuperscriptBox["10", 
           RowBox[{"-", "6"}]]}], "0", "0"},
        {"0", "0.001736", "0"},
        {"0", "0", "0.7417"}
       },
       GridBoxAlignment->{
        "Columns" -> {{Decimal}}, "ColumnsIndexed" -> {}, 
         "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}]}],
     ",", "\[IndentingNewLine]", 
    RowBox[{"RGEYd", "\[Rule]", 
     RowBox[{"(", "\[NoBreak]", GridBox[{
        {"0.0007314", 
         RowBox[{"0.001558", "\[InvisibleSpace]", "+", 
          RowBox[{"0.000022", " ", "\[ImaginaryI]"}]}], 
         RowBox[{"0.0002393", "\[InvisibleSpace]", "+", 
          RowBox[{"0.0006778", " ", "\[ImaginaryI]"}]}]},
        {
         RowBox[{"0.001555", "\[InvisibleSpace]", "-", 
          RowBox[{"0.000022", "\[ImaginaryI]"}]}], 
         RowBox[{"0.007555", "\[InvisibleSpace]"}], 
         RowBox[{"0.008913", "\[InvisibleSpace]", "-", 
          RowBox[{"0.000004", " ", "\[ImaginaryI]"}]}]},
        {
         RowBox[{"0.0002393", "\[InvisibleSpace]", "-", 
          RowBox[{"0.0006778", " ", "\[ImaginaryI]"}]}], 
         RowBox[{"0.008913", "\[InvisibleSpace]", "+", 
          RowBox[{"0.000004", " ", "\[ImaginaryI]"}]}], "0.2872"}
       },
       GridBoxAlignment->{
        "Columns" -> {{Decimal}}, "ColumnsIndexed" -> {}, 
         "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}]}],
     ",", "\[IndentingNewLine]", 
    RowBox[{"RGEYe", "\[Rule]", 
     RowBox[{"(", "\[NoBreak]", GridBox[{
        {"0.00009344", "0", "0"},
        {"0", "0.01972", "0"},
        {"0", "0", "0.33678"}
       },
       GridBoxAlignment->{
        "Columns" -> {{Decimal}}, "ColumnsIndexed" -> {}, 
         "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}]}],
     ",", "\[IndentingNewLine]", "\[IndentingNewLine]", 
    RowBox[{"RGEY\[Nu]", "\[Rule]", 
     RowBox[{"0.5", " ", 
      RowBox[{"(", "\[NoBreak]", GridBox[{
         {"0.01", "0", "0"},
         {
          RowBox[{"0.0001", "\[ImaginaryI]"}], "0.1", "0"},
         {"0", "0", "1"}
        },
        GridBoxAlignment->{
         "Columns" -> {{Decimal}}, "ColumnsIndexed" -> {}, 
          "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}], "\[NoBreak]", 
       ")"}]}]}], ",", "\[IndentingNewLine]", 
    RowBox[{"RGE\[Kappa]", "\[Rule]", 
     RowBox[{"(", "\[NoBreak]", GridBox[{
        {"0", "0", "0"},
        {"0", "0", "0"},
        {"0", "0", "0"}
       },
       GridBoxAlignment->{
        "Columns" -> {{Decimal}}, "ColumnsIndexed" -> {}, 
         "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}]}],
     ",", "\[IndentingNewLine]", 
    RowBox[{"RGEM\[Nu]r", "\[Rule]", 
     RowBox[{"(", "\[NoBreak]", GridBox[{
        {
         RowBox[{
          RowBox[{"-", "3.660"}], " ", 
          SuperscriptBox["10", "10"]}], 
         RowBox[{"1.675", " ", 
          SuperscriptBox["10", "10"]}], 
         RowBox[{
          RowBox[{"-", "1.675"}], " ", 
          SuperscriptBox["10", "11"]}]},
        {
         RowBox[{"1.675", " ", 
          SuperscriptBox["10", "10"]}], 
         RowBox[{
          RowBox[{"-", "2.218"}], " ", 
          SuperscriptBox["10", "12"]}], 
         RowBox[{"1.098", " ", 
          SuperscriptBox["10", "13"]}]},
        {
         RowBox[{
          RowBox[{"-", "1.675"}], " ", 
          SuperscriptBox["10", "11"]}], 
         RowBox[{"1.098", " ", 
          SuperscriptBox["10", "13"]}], 
         RowBox[{
          RowBox[{"-", "2.218"}], " ", 
          SuperscriptBox["10", "14"]}]}
       },
       GridBoxAlignment->{
        "Columns" -> {{Decimal}}, "ColumnsIndexed" -> {}, 
         "Rows" -> {{Baseline}}, "RowsIndexed" -> {}}], "\[NoBreak]", 
      ")"}]}]}], "\[IndentingNewLine]", "\[IndentingNewLine]", "]"}], 
  ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"RGESolve", "[", 
   RowBox[{"MZ", ",", "GUT"}], "]"}], ";"}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Plots", "Section"],

Cell[CellGroupData[{

Cell["Definitions", "Subsection"],

Cell["Define the plot region", "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"ClearAll", "[", 
   RowBox[{"tmin", ",", "tmax"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"tmin", "=", 
   RowBox[{"Log", "[", 
    RowBox[{"10", ",", "MZ"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"tmax", "=", 
   RowBox[{"Log", "[", 
    RowBox[{"10", ",", "GUT"}], "]"}]}], ";"}]}], "Input"],

Cell["Define the functions that should be plotted", "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"ClearAll", "[", 
   RowBox[{"Mnu", ",", "Ye"}], "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Mnu", "[", "mu_", "]"}], ":=", 
  RowBox[{
   RowBox[{"RGEGetSolution", "[", 
    RowBox[{"mu", ",", "RGEM\[Nu]"}], "]"}], "//", 
   "Chop"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Ye", "[", "mu_", "]"}], ":=", 
   RowBox[{
    RowBox[{"RGEGetSolution", "[", 
     RowBox[{"mu", ",", "RGEYe"}], "]"}], "//", "Chop"}]}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{"ClearAll", "[", 
  RowBox[{
  "MixingPars", ",", "MixingParsLogScale", ",", "NuMasses", ",", 
   "NuMassesLogScale"}], "]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"MixingPars", "[", "mu_", "]"}], ":=", 
  RowBox[{
   RowBox[{"MNSParameters", "[", 
    RowBox[{
     RowBox[{"Mnu", "[", "mu", "]"}], ",", 
     RowBox[{"Ye", "[", "mu", "]"}]}], "]"}], "\[LeftDoubleBracket]", "1", 
   "\[RightDoubleBracket]"}]}], "\n", 
 RowBox[{
  RowBox[{"MixingParsLogScale", "[", "x_", "]"}], ":=", 
  RowBox[{"MixingPars", "[", 
   RowBox[{"10", "^", "x"}], "]"}]}], "\n", 
 RowBox[{
  RowBox[{"NuMasses", "[", "mu_", "]"}], ":=", 
  RowBox[{
   RowBox[{"MNSParameters", "[", 
    RowBox[{
     RowBox[{"Mnu", "[", "mu", "]"}], ",", 
     RowBox[{"Ye", "[", "mu", "]"}]}], "]"}], "\[LeftDoubleBracket]", "2", 
   "\[RightDoubleBracket]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"NuMassesLogScale", "[", "x_", "]"}], ":=", 
   RowBox[{"NuMasses", "[", 
    RowBox[{"10", "^", "x"}], "]"}]}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{"ClearAll", "[", 
  RowBox[{
  "\[Theta]12", ",", "\[Theta]13", ",", "\[Theta]23", ",", "\[Delta]", ",", 
   "\[CurlyPhi]1", ",", "\[CurlyPhi]2", ",", "m1", ",", "m2", ",", "m3", ",", 
   "\[CapitalDelta]m2sol", ",", "\[CapitalDelta]m2atm"}], 
  "]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Theta]12", "[", "t_", "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"MixingParsLogScale", "[", "t", "]"}], "\[LeftDoubleBracket]", 
    "1", "\[RightDoubleBracket]"}], "/", "Degree"}]}], "\n", 
 RowBox[{
  RowBox[{"\[Theta]13", "[", "t_", "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"MixingParsLogScale", "[", "t", "]"}], "\[LeftDoubleBracket]", 
    "2", "\[RightDoubleBracket]"}], "/", "Degree"}]}], "\n", 
 RowBox[{
  RowBox[{"\[Theta]23", "[", "t_", "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"MixingParsLogScale", "[", "t", "]"}], "\[LeftDoubleBracket]", 
    "3", "\[RightDoubleBracket]"}], "/", 
   "Degree"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[Delta]", "[", "t_", "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"MixingParsLogScale", "[", "t", "]"}], "\[LeftDoubleBracket]", 
    "4", "\[RightDoubleBracket]"}], "/", "Degree"}]}], "\n", 
 RowBox[{
  RowBox[{"\[CurlyPhi]1", "[", "t_", "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"MixingParsLogScale", "[", "t", "]"}], "\[LeftDoubleBracket]", 
    "8", "\[RightDoubleBracket]"}], "/", 
   "Degree"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[CurlyPhi]2", "[", "t_", "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"MixingParsLogScale", "[", "t", "]"}], "\[LeftDoubleBracket]", 
    "9", "\[RightDoubleBracket]"}], "/", 
   "Degree"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"m1", "[", "t_", "]"}], ":=", 
  RowBox[{
   RowBox[{"NuMassesLogScale", "[", "t", "]"}], "\[LeftDoubleBracket]", "1", 
   "\[RightDoubleBracket]"}]}], "\n", 
 RowBox[{
  RowBox[{"m2", "[", "t_", "]"}], ":=", 
  RowBox[{
   RowBox[{"NuMassesLogScale", "[", "t", "]"}], "\[LeftDoubleBracket]", "2", 
   "\[RightDoubleBracket]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"m3", "[", "t_", "]"}], ":=", 
  RowBox[{
   RowBox[{"NuMassesLogScale", "[", "t", "]"}], "\[LeftDoubleBracket]", "3", 
   "\[RightDoubleBracket]"}]}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"\[CapitalDelta]m2sol", "[", "t_", "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"m2", "[", "t", "]"}], "^", "2"}], "-", 
   RowBox[{
    RowBox[{"m1", "[", "t", "]"}], "^", "2"}]}]}], "\n", 
 RowBox[{
  RowBox[{"\[CapitalDelta]m2atm", "[", "t_", "]"}], ":=", 
  RowBox[{
   RowBox[{
    RowBox[{"m3", "[", "t", "]"}], "^", "2"}], "-", 
   RowBox[{
    RowBox[{"m2", "[", "t", "]"}], "^", 
    "2"}]}]}], "\[IndentingNewLine]"}], "Input"],

Cell["\<\
Test for compatibility with the experimental data
(allowed regions at the 3\[Sigma] CL as given in Maltoni et al., \
hep-ph/0405172v4)\
\>", "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"ClearAll", "[", 
   RowBox[{"CompareToExperiment", ",", "IsCompatibleWithExp"}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"CompareToExperiment", "[", "t_", "]"}], ":=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"lNotComp", ",", "lComp"}], "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"lComp", "=", 
       RowBox[{"{", "}"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"lNotComp", "=", 
       RowBox[{"{", "}"}]}], ";", "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{"Sin", "[", 
            RowBox[{
             RowBox[{"\[Theta]12", "[", "t", "]"}], " ", "Degree"}], "]"}], 
           "^", "2"}], "\[GreaterEqual]", "0.23"}], "&&", 
         RowBox[{
          RowBox[{
           RowBox[{"Sin", "[", 
            RowBox[{
             RowBox[{"\[Theta]12", "[", "t", "]"}], " ", "Degree"}], "]"}], 
           "^", "2"}], "\[LessEqual]", "0.38"}]}], ",", 
        RowBox[{"lComp", "=", 
         RowBox[{"Append", "[", 
          RowBox[{"lComp", ",", "\"\<\[Theta]12\>\""}], "]"}]}], ",", 
        RowBox[{"lNotComp", "=", 
         RowBox[{"Append", "[", 
          RowBox[{"lNotComp", ",", "\"\<\[Theta]12\>\""}], "]"}]}]}], "]"}], 
      ";", "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"Sin", "[", 
           RowBox[{
            RowBox[{"\[Theta]13", "[", "t", "]"}], " ", "Degree"}], "]"}], 
          "^", "2"}], "\[LessEqual]", "0.051"}], ",", 
        RowBox[{"lComp", "=", 
         RowBox[{"Append", "[", 
          RowBox[{"lComp", ",", "\"\<\[Theta]13\>\""}], "]"}]}], ",", 
        RowBox[{"lNotComp", "[", 
         RowBox[{"lNotComp", ",", "\"\<\[Theta]13\>\""}], "]"}]}], "]"}], ";",
       "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{"Sin", "[", 
            RowBox[{
             RowBox[{"\[Theta]23", "[", "t", "]"}], " ", "Degree"}], "]"}], 
           "^", "2"}], "\[GreaterEqual]", "0.34"}], "&&", 
         RowBox[{
          RowBox[{
           RowBox[{"Sin", "[", 
            RowBox[{
             RowBox[{"\[Theta]23", "[", "t", "]"}], " ", "Degree"}], "]"}], 
           "^", "2"}], "\[LessEqual]", "0.68"}]}], ",", 
        RowBox[{"lComp", "=", 
         RowBox[{"Append", "[", 
          RowBox[{"lComp", ",", "\"\<\[Theta]23\>\""}], "]"}]}], ",", 
        RowBox[{"lNotComp", "=", 
         RowBox[{"Append", "[", 
          RowBox[{"lNotComp", ",", "\"\<\[Theta]23\>\""}], "]"}]}]}], "]"}], 
      ";", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"Abs", "[", 
           RowBox[{"\[CapitalDelta]m2atm", "[", "t", "]"}], "]"}], 
          "\[GreaterEqual]", 
          RowBox[{"1.3", " ", 
           RowBox[{"10", "^", 
            RowBox[{"-", "3"}]}]}]}], "&&", 
         RowBox[{
          RowBox[{"Abs", "[", 
           RowBox[{"\[CapitalDelta]m2atm", "[", "t", "]"}], "]"}], 
          "\[LessEqual]", 
          RowBox[{"3.2", " ", 
           RowBox[{"10", "^", 
            RowBox[{"-", "3"}]}]}]}]}], ",", 
        RowBox[{"lComp", "=", 
         RowBox[{"Append", "[", 
          RowBox[{"lComp", ",", "\"\<\[CapitalDelta]m2atm\>\""}], "]"}]}], 
        ",", 
        RowBox[{"lNotComp", "=", 
         RowBox[{"Append", "[", 
          RowBox[{"lNotComp", ",", "\"\<\[CapitalDelta]m2atm\>\""}], 
          "]"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"\[CapitalDelta]m2sol", "[", "t", "]"}], "\[GreaterEqual]", 
          
          RowBox[{"7.1", " ", 
           RowBox[{"10", "^", 
            RowBox[{"-", "5"}]}]}]}], "&&", 
         RowBox[{
          RowBox[{"\[CapitalDelta]m2sol", "[", "t", "]"}], "\[LessEqual]", 
          RowBox[{"8.9", " ", 
           RowBox[{"10", "^", 
            RowBox[{"-", "5"}]}]}]}]}], ",", 
        RowBox[{"lComp", "=", 
         RowBox[{"Append", "[", 
          RowBox[{"lComp", ",", "\"\<\[CapitalDelta]m2sol\>\""}], "]"}]}], 
        ",", 
        RowBox[{"lNotComp", "=", 
         RowBox[{"Append", "[", 
          RowBox[{"lNotComp", ",", "\"\<\[CapitalDelta]m2sol\>\""}], 
          "]"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"Print", "[", 
       RowBox[{
       "\"\<The following parameters are compatible with the experimental \
data at the 3\[Sigma] CL:\>\"", ",", "lComp"}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"Length", "[", "lNotComp", "]"}], "\[Equal]", "0"}], ",", 
        RowBox[{"Return", "[", "1", "]"}], ",", 
        RowBox[{"Return", "[", "0", "]"}]}], "]"}], ";"}]}], 
    "\[IndentingNewLine]", "]"}]}], ";"}], "\n", 
 RowBox[{"IsCompatibleWithExp", ":=", 
  RowBox[{"CompareToExperiment", "[", 
   RowBox[{"Log", "[", 
    RowBox[{"10", ",", "MZ"}], "]"}], "]"}]}]}], "Input"],

Cell["Positions of the thresholds", "Text"],

Cell[BoxData[{
 RowBox[{"ClearAll", "[", 
  RowBox[{"M1", ",", "M2", ",", "M3", ",", "ShadowEFT"}], "]"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"{", 
    RowBox[{"M1", ",", "M2", ",", "M3"}], "}"}], "=", 
   RowBox[{
    RowBox[{"RGEGetTransitions", "[", "]"}], "\[LeftDoubleBracket]", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"4", ",", "3", ",", "2"}], "}"}], ",", "1"}], 
    "\[RightDoubleBracket]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{"ShadowEFT", "=", 
   RowBox[{"RGEShadowEFT", "[", 
    RowBox[{"M1", ",", "M2", ",", "M3", ",", 
     SuperscriptBox["10", "17"]}], "]"}]}], ";"}]}], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compatibility to the experimental data", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"IsCompatibleWithExp", ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The following parameters are compatible with the experimental \
data at the 3\[Sigma] CL:\"\>", "\[InvisibleSpace]", 
   RowBox[{"{", 
    RowBox[{"\<\"\[Theta]12\"\>", ",", "\<\"\[Theta]13\"\>", 
     ",", "\<\"\[Theta]23\"\>", ",", "\<\"\[CapitalDelta]m2atm\"\>", 
     ",", "\<\"\[CapitalDelta]m2sol\"\>"}], "}"}]}],
  SequenceForm[
  "The following parameters are compatible with the experimental data at the \
3\[Sigma] CL:", {
   "\[Theta]12", "\[Theta]13", "\[Theta]23", "\[CapitalDelta]m2atm", 
    "\[CapitalDelta]m2sol"}],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847119701901*^9, 3.636847462731737*^9, 
  3.646456138380522*^9, 3.646475887374419*^9, 3.646477146195882*^9, 
  3.6464777994864683`*^9}]
}, Open  ]],

Cell["The low energy parameters are", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<\!\(\*SubscriptBox[\(\[Theta]\), \(12\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\))=\>\"", ",", 
    RowBox[{"\[Theta]12", "[", 
     RowBox[{"Log", "[", 
      RowBox[{"10", ",", "MZ"}], "]"}], "]"}], ",", "\"\<\[Degree]\>\""}], 
   "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<\!\(\*SubscriptBox[\(\[Theta]\), \(13\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\))=\>\"", ",", 
    RowBox[{"\[Theta]13", "[", 
     RowBox[{"Log", "[", 
      RowBox[{"10", ",", "MZ"}], "]"}], "]"}], ",", "\"\<\[Degree]\>\""}], 
   "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<\!\(\*SubscriptBox[\(\[Theta]\), \(23\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\))=\>\"", ",", 
    RowBox[{"\[Theta]23", "[", 
     RowBox[{"Log", "[", 
      RowBox[{"10", ",", "MZ"}], "]"}], "]"}], ",", "\"\<\[Degree]\>\""}], 
   "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<\!\(\*SubscriptBox[SuperscriptBox[\(\[CapitalDelta]m\), \(2\)], \(atm\
\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\))=\>\"", ",", 
    RowBox[{"\[CapitalDelta]m2atm", "[", 
     RowBox[{"Log", "[", 
      RowBox[{"10", ",", "MZ"}], "]"}], "]"}], ",", 
    "\"\<\!\(\*SuperscriptBox[\(eV\), \(2\)]\)\>\""}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<\!\(\*SubscriptBox[SuperscriptBox[\(\[CapitalDelta]m\), \(2\)], \(sol\
\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\))=\>\"", ",", 
    RowBox[{"\[CapitalDelta]m2sol", "[", 
     RowBox[{"Log", "[", 
      RowBox[{"10", ",", "MZ"}], "]"}], "]"}], ",", 
    "\"\<\!\(\*SuperscriptBox[\(eV\), \(2\)]\)\>\""}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<\!\(\*SubscriptBox[\(m\), \(1\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\))=\>\"", ",", 
    RowBox[{"m1", "[", 
     RowBox[{"Log", "[", 
      RowBox[{"10", ",", "MZ"}], "]"}], "]"}], ",", "\"\<eV\>\""}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<\!\(\*SubscriptBox[\(m\), \(2\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\))=\>\"", ",", 
    RowBox[{"m2", "[", 
     RowBox[{"Log", "[", 
      RowBox[{"10", ",", "MZ"}], "]"}], "]"}], ",", "\"\<eV\>\""}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<\!\(\*SubscriptBox[\(m\), \(3\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\))=\>\"", ",", 
    RowBox[{"m3", "[", 
     RowBox[{"Log", "[", 
      RowBox[{"10", ",", "MZ"}], "]"}], "]"}], ",", "\"\<eV\>\""}], "]"}], 
  ";"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"\\!\\(\\*SubscriptBox[\\(\[Theta]\\), \
\\(12\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\))=\"\>", 
   "\[InvisibleSpace]", "32.494768972620406`", 
   "\[InvisibleSpace]", "\<\"\[Degree]\"\>"}],
  SequenceForm[
  "\!\(\*SubscriptBox[\(\[Theta]\), \(12\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\))=", 32.494768972620406`, "\[Degree]"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847119716612*^9, 3.636847462760044*^9, 
  3.646456138430859*^9, 3.646475887409971*^9, 3.646477146277935*^9, 
  3.6464777995207*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"\\!\\(\\*SubscriptBox[\\(\[Theta]\\), \
\\(13\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\))=\"\>", 
   "\[InvisibleSpace]", "0.10085307393552177`", 
   "\[InvisibleSpace]", "\<\"\[Degree]\"\>"}],
  SequenceForm[
  "\!\(\*SubscriptBox[\(\[Theta]\), \(13\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\))=", 0.10085307393552177`, "\[Degree]"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847119716612*^9, 3.636847462760044*^9, 
  3.646456138430859*^9, 3.646475887409971*^9, 3.646477146277935*^9, 
  3.646477799521927*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"\\!\\(\\*SubscriptBox[\\(\[Theta]\\), \
\\(23\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\))=\"\>", 
   "\[InvisibleSpace]", "46.96199065442854`", 
   "\[InvisibleSpace]", "\<\"\[Degree]\"\>"}],
  SequenceForm[
  "\!\(\*SubscriptBox[\(\[Theta]\), \(23\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\))=", 46.96199065442854, "\[Degree]"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847119716612*^9, 3.636847462760044*^9, 
  3.646456138430859*^9, 3.646475887409971*^9, 3.646477146277935*^9, 
  3.646477799559121*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"\\!\\(\\*SubscriptBox[SuperscriptBox[\\(\[CapitalDelta]m\\), \
\\(2\\)], \\(atm\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\))=\"\>", 
   "\[InvisibleSpace]", "0.002092995808760806`", 
   "\[InvisibleSpace]", "\<\"\\!\\(\\*SuperscriptBox[\\(eV\\), \
\\(2\\)]\\)\"\>"}],
  SequenceForm[
  "\!\(\*SubscriptBox[SuperscriptBox[\(\[CapitalDelta]m\), \(2\)], \
\(atm\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\))=", 0.002092995808760806, 
   "\!\(\*SuperscriptBox[\(eV\), \(2\)]\)"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847119716612*^9, 3.636847462760044*^9, 
  3.646456138430859*^9, 3.646475887409971*^9, 3.646477146277935*^9, 
  3.64647779957298*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"\\!\\(\\*SubscriptBox[SuperscriptBox[\\(\[CapitalDelta]m\\), \
\\(2\\)], \\(sol\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\))=\"\>", 
   "\[InvisibleSpace]", "0.00007720436037501739`", 
   "\[InvisibleSpace]", "\<\"\\!\\(\\*SuperscriptBox[\\(eV\\), \
\\(2\\)]\\)\"\>"}],
  SequenceForm[
  "\!\(\*SubscriptBox[SuperscriptBox[\(\[CapitalDelta]m\), \(2\)], \
\(sol\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\))=", 0.00007720436037501739, 
   "\!\(\*SuperscriptBox[\(eV\), \(2\)]\)"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847119716612*^9, 3.636847462760044*^9, 
  3.646456138430859*^9, 3.646475887409971*^9, 3.646477146277935*^9, 
  3.646477799574793*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"\\!\\(\\*SubscriptBox[\\(m\\), \
\\(1\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\))=\"\>", 
   "\[InvisibleSpace]", "0.014843437266616027`", 
   "\[InvisibleSpace]", "\<\"eV\"\>"}],
  SequenceForm[
  "\!\(\*SubscriptBox[\(m\), \(1\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\))=", 
   0.014843437266616027`, "eV"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847119716612*^9, 3.636847462760044*^9, 
  3.646456138430859*^9, 3.646475887409971*^9, 3.646477146277935*^9, 
  3.646477799576034*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"\\!\\(\\*SubscriptBox[\\(m\\), \
\\(2\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\))=\"\>", 
   "\[InvisibleSpace]", "0.017249115637127107`", 
   "\[InvisibleSpace]", "\<\"eV\"\>"}],
  SequenceForm[
  "\!\(\*SubscriptBox[\(m\), \(2\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\))=", 
   0.017249115637127107`, "eV"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847119716612*^9, 3.636847462760044*^9, 
  3.646456138430859*^9, 3.646475887409971*^9, 3.646477146277935*^9, 
  3.646477799577059*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"\\!\\(\\*SubscriptBox[\\(m\\), \
\\(3\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\))=\"\>", 
   "\[InvisibleSpace]", "0.0488930240323074`", 
   "\[InvisibleSpace]", "\<\"eV\"\>"}],
  SequenceForm[
  "\!\(\*SubscriptBox[\(m\), \(3\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\))=", 
   0.0488930240323074, "eV"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847119716612*^9, 3.636847462760044*^9, 
  3.646456138430859*^9, 3.646475887409971*^9, 3.646477146277935*^9, 
  3.646477799578034*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Mixing Angles", "Subsection"],

Cell["Draw the plot for the mixing angles", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"Plot", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"\[Theta]12", "[", "t", "]"}], ",", 
       RowBox[{"\[Theta]13", "[", "t", "]"}], ",", 
       RowBox[{"\[Theta]23", "[", "t", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "tmin", ",", "tmax"}], "}"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"ImageSize", "\[Rule]", "400"}], ",", 
     RowBox[{"FrameLabel", "->", 
      RowBox[{"{", 
       RowBox[{
       "\"\<\!\(\*SubscriptBox[\(log\), \(10\)]\)(\[Mu]/GeV)\>\"", ",", 
        "\"\<\>\""}], "}"}]}], ",", 
     RowBox[{"FrameTicks", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"RGELogTicksLabeled", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"i", "*", "15"}], ",", 
            RowBox[{
             RowBox[{"ToString", "[", 
              RowBox[{"i", "*", "15"}], "]"}], "<>", "\"\<\[Degree]\>\""}]}], 
           "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", "0", ",", "3"}], "}"}]}], "]"}], ",", 
        RowBox[{"RGELogTicks", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"i", "*", "15"}], ",", "\"\<\>\""}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", "0", ",", "3"}], "}"}]}], "]"}]}], "}"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"PlotRange", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"All", ",", "All"}], "}"}]}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"Red", ",", "ForestGreen", ",", "Blue"}], "}"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"Prolog", "\[Rule]", 
      RowBox[{"{", "ShadowEFT", "}"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"Epilog", "\[Rule]", 
      RowBox[{"{", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"StyleForm", "[", 
           RowBox[{
            SubscriptBox["\[Theta]", "12"], ",", 
            RowBox[{"FontColor", "\[Rule]", "Red"}], ",", 
            RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"tmin", ",", 
            RowBox[{"\[Theta]12", "[", "tmin", "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "0.8"}], ",", "1.4"}], "}"}]}], "]"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"StyleForm", "[", 
           RowBox[{
            SubscriptBox["\[Theta]", "13"], ",", 
            RowBox[{"FontColor", "\[Rule]", "ForestGreen"}], ",", 
            RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"tmin", ",", 
            RowBox[{"\[Theta]13", "[", "tmin", "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "0.8"}], ",", 
            RowBox[{"-", "1"}]}], "}"}]}], "]"}], ",", "\[IndentingNewLine]", 
        
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"StyleForm", "[", 
           RowBox[{
            SubscriptBox["\[Theta]", "23"], ",", 
            RowBox[{"FontColor", "\[Rule]", "Blue"}], ",", 
            RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"tmin", ",", 
            RowBox[{"\[Theta]23", "[", "tmin", "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "0.8"}], ",", "1.4"}], "}"}]}], "]"}]}], "}"}]}], 
     ",", "\[IndentingNewLine]", 
     RowBox[{"DisplayFunction", "\[Rule]", "Identity"}]}], 
    "\[IndentingNewLine]", "]"}], ",", 
   RowBox[{"DisplayFunction", "\[Rule]", "$DisplayFunction"}]}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.435655087000834*^9, 3.435655149172597*^9}, {
   3.435655291841278*^9, 3.435655355632481*^9}, {3.435655474900189*^9, 
   3.43565552545153*^9}, 3.435655622860778*^9}],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {RGBColor[1, 0, 0], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwd1XlcTekDx/HWe297JEIqKW0nXdOm7rn3fhGtKgpTriZLymjTTxgtokV+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     "]]}, 
   {RGBColor[0.1, 0.75, 0.2], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwV03k0VV0YBvAI94oyRKMphDjuOZIU8b5JKSpDfRWVDCmVEmXIVFQSJYWo
cA+5RJkLtxShUqgoU6FINCCUISXf8cdee/3W3mvt5zzvOouc3G1cBKdNm9bF
rKm92SN7eYzdJPzUN3kZXMeHMLMHtvFuk/C+alwqhLG+fFVAQtAksOyK/pvy
1aqO8tSUSbA+uqptylsVZS3vfZuEs2u9e08zdu1+qHvIaBrO000XC2XcFa26
8ULtNHz2bWhjOOP3/YMBL/sE8NWy1ndXGHdv4S2rsxdChyef43iM9bxkon/s
F8LfR+zap3zm6unBmR5CKOder5rGeFGLY7b5aSHc7FKTN+Vd9goalelCuGbG
l+p0xm9c4xYWDQjhIemH0zIZlwWeF0wMEcb2TQ1u2YwluGMODyKEMVwgv2DK
9mX7S1tihFG88/L4lP8IrguYky6MJ+R3huUwXh4mOHKpWhi3u85OzWWcdcX/
a4iMCI4YLmjNZ3w9/fBr1zQRjFjlu7mI8bi33V7bHBFco2MTM2XbdWZjG4pF
0ItNvpvyvM9KSlovRHCb26BLMeNY5bfufX0iKPA2PpDPOCrRQMJTj4XSmh5Z
DxgPuKnfdDJmPP3EzylbrpLR32LGwp8Fp1eWMJ71vs9+mS3ji8lPphwxLzl7
JICFBhKTbQ8Zf+u5aNJzloUtXvKqjxibF/k1NkWycCjN+OCU2du2/iumWTh6
+PzIlEOjWZv9K1kIe1CibGo+Tr86DtWycEDc678pr13a4bWrkYWlgXduTDlV
4KXopg/M/ZzujikL1d1PNPrCQv1UFY3HjJ3pdB3OIAsl7JyPTLncPeaJwjgL
UxwXLJ+6rwzBthLT2Sied41czDh41pG+STE2St7tct75ig8f2+yCB2TYaCJ7
O/FyLR8wy0y2Q56NuuIDrc+q+fDPYpFROcnGuFhBt2VVfAg4//bCmS1snL0k
tWbmYz607nis5LWLjQtM+UtNS/lgqJF918WFjZX5+5L8HjLzenqudZ0PG6vc
4kJ7+HzwFTIkRG+wcU1C5vXH+Xw4djK5+mIn8/7Lg2+9U/kQmde9A3vZ6HIj
9G1oCh8yP2l1Dw2z8WL/8fdXaT50mhVO2yEqivNzWiYKE/hgI1mtp0yJ4q8d
+rGjMXygkn9xCwNF8cbcfim/s3zoLzc7/mHeDBzwR6cz+5j/R6RfTsdSDN/o
aK/K0eCDSIClzYErM9Hd8XnT9XvF4PC6q19iuiS2D4ZNHF5ZDK9St3fp6Evh
BGe0VbGsCKIH3lokbZPGuvLSvVGWRfA52r59Q+RsVDCpqdBrLISI7KFOyJNB
JQOZCdqlEEwjVsUFvJTFotVKlr9H70FE6RXvselzET69vdMcfA/UDv189kh1
HrbsC7zvI3MPLI84mFbuno+lFSc0OpLuQoTu6aOSlxdgVqChhT9xF+SO5s5K
z1mI9elik5XlBWAg9K0poFsOhWtTNhhaF0CWBGd2IlsBhQsOzNr5JR+yPYNO
/DVRxENyjmajPvnQYVFuVmeohJL6ND04Ix9kr7+IeF6thFvEhIJ80vJAUsbW
pWP/IqxTs/nQa5AH9573+iT9XoQDOhY7OppyYa/Dzs6gMGV0aDVRXeibC++U
rTqDVFTw8OjVzEdSuTCuZL5Q9a4K6q0JX3F0fw50mL/4+sVKFSVdB1sFudlQ
999iybpvqjj/5njdi7YscEs4627vtxilFB6ueE9kQe0st71ic9Twl7k7mXP0
DrSUbL+yO10N3dc6dnArb0O/hcK8N3rqOJy2Otxd8Ta8EDmk6/paHcl1VY36
RzPBaqLkXIqDBia324ykvM6ARMVYj79/NbD8hrVW15IMiA30jPkTuQS/2/pE
Jly5BVdLNCUUFDRxaOGucjvBW2CvfkDRga+JX9PCXl84mA6va++4mplr4d5o
s0zHjjQQpb3NNvdoofj0x3rV1mnQuNR6oPCrFo79Kp9MtkyD1Cda1xR7tTDf
a778iU1pYPz947eBAS0sGOyx1tiQBp7LLSKjx7XQlSvMC8c0eFej2NA8k8Cj
hPxvezINMn9XOTkvI9BfoCBu3cw08LlwU6xmOYEEZrepi6WBqWLQ3WUrCYwS
U/AVFU2DdlNdERFjAl1DvTJrhdJAOiox45YZgT8zW4Tt/vLAT81zoM+OwAsT
wtyQ7zyw2LIwyDeYwMH4mLjZL3gg8qflQdFpAtNEOperVPGgPCX+98hZAgWi
rinpPuWBwZCsl1c4gZVznd5sKeeBxmVJN89oAmcoyhXF3+dB14pXGXmxBOp2
+GdmFvOA/nixZyCOwNxNgV9KCnkwlxJzdk8gcKZ0sGlnPg/qm57T2UlMXpvL
TsO5PIg8GdbeRxM4Nm9PNTuHyfNK2M6NR+DkiQoz6jaTx7sy7nY6gcdKI96Y
ZvAgSOF0w7cMAvet/LHHNp3J93T1bM07BJ4Myj2/PZUHw4enWR/IJlDhzERL
VhIPcmVLI2/lEqgynvlJ+BoP3B4G1vTkE3j/3Pxbe6KZ/OJ/zPYVESjR83jf
vDAm/13+WR6fQO8D5c4+ITzYtcu3ousBgUkaUouaA3jw5vawsXMZgf0zdDqT
PJh85gN6Ds8JLH/2jrd6J5NvKPsYt5pA41xeTfZ/TL7rh/Paa5nzMnapvBUP
8r5+I3bXE2gVYykusJZ5L+yzit17Zh5FSh5tHB4EPn0ntbWPwIOCKhn9gjxY
YfKkz0xKG+dKKi8NvJMKvbuGh5ZnaaO9uKNRyaabsG5p/5rAHG08WO4W8GP9
TeCyemIq8rRRLWem4WLTm2CT37zcqlAb30pdNo43uAnFrBK/A6XamNGzOzpB
/SacyQ8RTKjTRg2tGjklwZuwkC0pLTCijcYm55cp8VPAvEBTp9qYg0Wbu+55
kimg8StEQmQ1B4dXcloOaKWAsN77PlzDwT05zq3O6ilQVhiRWWjGQdlJCHdQ
TAH9+70qKVbMuWkcP0AiBVQfZ8854cRBFtdijslAMvx7ufSvWigHB4eCLWfd
TYZWiYh3jmEcfKPwkHDJTQa+1afihHAOKmtlqZfeSYbj9dFe0pc4KLplf1AA
Lxm+N/zqn4jj4O46+89yccnQ3FrY8TaDg4rZKRtG/JIh75vBs5BaDl7Nc/44
b20yGJt+Vel/xcE/H9+ciVqdDC8S407Z1nMw+32enLhxMnRZ/VpJNXEwStqB
ktBPhjlFWXfaPnIwxHZIc+2SZAg4rXRlxS8O6i2xOus5KxnWy4ns7l9A4tZW
KdtP72lo8LrLt5UnEaZ/FG5vpsHpldOcJ4okrlxmM9jaQIN/SOmrG6okvonl
XOx5RUP2F1+TDRwS+78Wz6UqaZC59109dTWJu34+PxCVRcOHTXVDtq4klouV
aCSepsGuxLnrxkESN8gskN8cTEOj5khDmxuJF+U2TAqepKGGtZDv6EHi5dH4
aF8/Gu4/3nvygB+JRAF7SbQHDVf1xsROXCCRY1bues6BBunU8In7kSS23W0d
87Gn4aK0/I+/USRevdVwyW0XDWd+rK4/FUvioMzOZscdNHhkRMSHJZG4IPPP
3wArGjbKKy6OzyXxztC+Ju/VNFRF5M15l09iZ3DL6VtAw5rxNWy5eyRKLWVt
bjeiwaDJ9TuXT2JK0CBuM6BB/XJ+Xno5iS7sf098dGkQFF5nXPSWxFx/CbcQ
NRqCjjWTvxtJ1FVOvzSiSsPvjoOLDFtIrDJf8fmICg0Djy4JlbaRuGPtsKab
Eg1tvi0vnnaTGMwdx2sLaCjuddveOEbiQYPQyEcSNPgJ2Zza+YdEad33nlaz
aDCS08/4OEHiuasb03vEaejgnQr0EqAwvmJUWG0GDfsazF+dFKHwt1VIaacQ
DWaOD53LWRRe3/hjA3c6DRp95JiQKIVRUe0cR0EavgvJLgoXoxBjyQXfJrlw
dNkHzxhJCq0H1vrcH+GCdZk1q0mKwgt7u73O/+LC0o2VN+bPprAn9rWh8xAX
hp0zKpNkKez/8WGxfj8X/KKPzclYQOHxGzq3Hbq5sEuxO/P7QgoX9+XIhXZx
wej2DuDIU6hLKrsUdnJhWoWRa4EihX3SvcH6H7hwdoh1/6EqhdH2Y/NLmrmw
L8hv8zQ1CsXDTD1XNHFh/Yy+ThN1ChujA7JLG7gwQ7lerGoJhUJjpF5HPRd6
s0zpGVoUFsi/Tj5fx4XalUXLNhEUupe0jq56zYUo64Td9Rzme/8tdams5YJH
26whGYrCpGOtQfE1XLA5EBy6XYfCHxN5oT7VXJAN3pfTqkthwrHFLtuec2FE
vGWNoh6F8gbPjbZVcaE53qLZcTmFR6tT2Y7PuHA9lxLoWUHh6PRo37gnXPBf
dTN2iQGFIhNWihWVTB9VsppuhhSm+yY++F3BBeOtYY+yV1GolaCz2ZCx4sdx
m0EjCpW1HzSdK2f6cTvcowsU7j/ZuuPDYy50jn7w90YKOe5E/WrGqRJPUv+Y
UHjSxS9HvYzp74b+SmNTCm81usncLmX6U8+sPbWWQpMy72MrGa8vkHOqWEfh
A+5wTf0jLiyBSyPC6yk86GG4yIfxjGqBiPUbKDTOijyqxrh323HFCHMKJdZp
Pvj4kOmzs7ug1oJCs1tGAmmMs4/YrpfcxOTn95p4M/4fVcmgyQ==
     "]]}, 
   {RGBColor[0, 0, 1], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwV03s4lFkcB3A0Zka3cZvWpoyYdZl55bW5tHL5rez7uqxcV43QFJUyu1pC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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{False, False},
  AxesLabel->{None, None},
  AxesOrigin->{2., 0},
  DisplayFunction->Identity,
  Epilog->{
    InsetBox[
     FormBox[
      TagBox[
       StyleBox[
        SubscriptBox["\[Theta]", "12"], FontColor -> RGBColor[1, 0, 0], 
        FontSize -> 12], 
       StyleForm[#, FontColor -> RGBColor[1, 0, 0], FontSize -> 12]& ], 
      TraditionalForm], {1.9596137110710439`, 32.494768972620406`}, {-0.8, 
     1.4}], 
    InsetBox[
     FormBox[
      TagBox[
       StyleBox[
        SubscriptBox["\[Theta]", "13"], FontColor -> RGBColor[0.1, 0.75, 0.2],
         FontSize -> 12], 
       StyleForm[#, FontColor -> RGBColor[0.1, 0.75, 0.2], FontSize -> 12]& ],
       TraditionalForm], {1.9596137110710439`, 
     0.10085307393552177`}, {-0.8, -1}], 
    InsetBox[
     FormBox[
      TagBox[
       StyleBox[
        SubscriptBox["\[Theta]", "23"], FontColor -> RGBColor[0, 0, 1], 
        FontSize -> 12], 
       StyleForm[#, FontColor -> RGBColor[0, 0, 1], FontSize -> 12]& ], 
      TraditionalForm], {1.9596137110710439`, 46.96199065442854}, {-0.8, 
     1.4}]},
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{
     FormBox["\"\"", TraditionalForm], None}, {
     FormBox[
     "\"\\!\\(\\*SubscriptBox[\\(log\\), \\(10\\)]\\)(\[Mu]/GeV)\"", 
      TraditionalForm], None}},
  FrameTicks->{{{{0, 
       FormBox["\"0\[Degree]\"", TraditionalForm]}, {15, 
       FormBox["\"15\[Degree]\"", TraditionalForm]}, {30, 
       FormBox["\"30\[Degree]\"", TraditionalForm]}, {45, 
       FormBox["\"45\[Degree]\"", TraditionalForm]}}, {{0, 
       FormBox["\"\"", TraditionalForm]}, {15, 
       FormBox["\"\"", TraditionalForm]}, {30, 
       FormBox["\"\"", TraditionalForm]}, {45, 
       FormBox["\"\"", TraditionalForm]}}}, {{{2, 
       FormBox["2", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["3", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["4", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["5", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["6", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["7", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["8", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["9", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["10", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["11", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["12", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["13", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["14", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["15", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["16", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["17", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["18", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["19", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}, {{2, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}}},
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImageSize->400,
  Method->{"DefaultBoundaryStyle" -> Automatic, "ScalingFunctions" -> None},
  PlotLabel->FormBox["\"\"", TraditionalForm],
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.05], 
     Scaled[0.05]}},
  Prolog->{{{
      GrayLevel[0.9], 
      RectangleBox[{10.561453473142157`, -10000}, {
       12.222494362608364`, 10000}]}, {
      GrayLevel[0.8], 
      RectangleBox[{12.222494362608364`, -10000}, {
       14.334981351900138`, 10000}]}, {
      GrayLevel[0.7], 
      RectangleBox[{14.334981351900138`, -10000}, {17, 10000}]}}},
  Ticks->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.6368471205099077`*^9, 3.636847463561467*^9, 
  3.646456139320549*^9, 3.646475888383235*^9, 3.6464771476670103`*^9, 
  3.6464778010481577`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["RG change of the angles:", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<The RG change of \!\(\*SubscriptBox[\(\[Theta]\), \(12\)]\) is :    \
\!\(\*SubscriptBox[\(\[Theta]\), \(12\)]\)(\!\(\*SubscriptBox[\(M\), \(GUT\)]\
\))-\!\(\*SubscriptBox[\(\[Theta]\), \(12\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\)) = \>\"", ",", 
    RowBox[{
     RowBox[{"\[Theta]12", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "GUT"}], "]"}], "]"}], "-", 
     RowBox[{"\[Theta]12", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "MZ"}], "]"}], "]"}]}], ",", "\"\<\[Degree]\>\""}], 
   "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<The RG change of \!\(\*SubscriptBox[\(\[Theta]\), \(13\)]\) is :    \
\!\(\*SubscriptBox[\(\[Theta]\), \(13\)]\)(\!\(\*SubscriptBox[\(M\), \(GUT\)]\
\))-\!\(\*SubscriptBox[\(\[Theta]\), \(13\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\)) = \>\"", ",", 
    RowBox[{
     RowBox[{"\[Theta]13", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "GUT"}], "]"}], "]"}], "-", 
     RowBox[{"\[Theta]13", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "MZ"}], "]"}], "]"}]}], ",", "\"\<\[Degree]\>\""}], 
   "]"}], ";", 
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<The RG change of \!\(\*SubscriptBox[\(\[Theta]\), \(23\)]\) is :    \
\!\(\*SubscriptBox[\(\[Theta]\), \(23\)]\)(\!\(\*SubscriptBox[\(M\), \(GUT\)]\
\))-\!\(\*SubscriptBox[\(\[Theta]\), \(23\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\)) = \>\"", ",", 
    RowBox[{
     RowBox[{"\[Theta]23", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "GUT"}], "]"}], "]"}], "-", 
     RowBox[{"\[Theta]23", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "MZ"}], "]"}], "]"}]}], ",", "\"\<\[Degree]\>\""}], 
   "]"}], ";"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The RG change of \\!\\(\\*SubscriptBox[\\(\[Theta]\\), \
\\(12\\)]\\) is :    \\!\\(\\*SubscriptBox[\\(\[Theta]\\), \
\\(12\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \
\\(GUT\\)]\\))-\\!\\(\\*SubscriptBox[\\(\[Theta]\\), \
\\(12\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\)) = \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"-", "5.486653027101081`"}], 
   "\[InvisibleSpace]", "\<\"\[Degree]\"\>"}],
  SequenceForm[
  "The RG change of \!\(\*SubscriptBox[\(\[Theta]\), \(12\)]\) is :    \
\!\(\*SubscriptBox[\(\[Theta]\), \(12\)]\)(\!\(\*SubscriptBox[\(M\), \(GUT\)]\
\))-\!\(\*SubscriptBox[\(\[Theta]\), \(12\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\)) = ", -5.486653027101081, "\[Degree]"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847120664076*^9, 3.6368474637158937`*^9, 
  3.646456139612596*^9, 3.646475888647488*^9, 3.646477147918311*^9, 
  3.646477801164693*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The RG change of \\!\\(\\*SubscriptBox[\\(\[Theta]\\), \
\\(13\\)]\\) is :    \\!\\(\\*SubscriptBox[\\(\[Theta]\\), \
\\(13\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \
\\(GUT\\)]\\))-\\!\\(\\*SubscriptBox[\\(\[Theta]\\), \
\\(13\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\)) = \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"-", "0.06985689645721001`"}], 
   "\[InvisibleSpace]", "\<\"\[Degree]\"\>"}],
  SequenceForm[
  "The RG change of \!\(\*SubscriptBox[\(\[Theta]\), \(13\)]\) is :    \
\!\(\*SubscriptBox[\(\[Theta]\), \(13\)]\)(\!\(\*SubscriptBox[\(M\), \(GUT\)]\
\))-\!\(\*SubscriptBox[\(\[Theta]\), \(13\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\)) = ", -0.06985689645721001, "\[Degree]"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847120664076*^9, 3.6368474637158937`*^9, 
  3.646456139612596*^9, 3.646475888647488*^9, 3.646477147918311*^9, 
  3.646477801168137*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The RG change of \\!\\(\\*SubscriptBox[\\(\[Theta]\\), \
\\(23\\)]\\) is :    \\!\\(\\*SubscriptBox[\\(\[Theta]\\), \
\\(23\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \
\\(GUT\\)]\\))-\\!\\(\\*SubscriptBox[\\(\[Theta]\\), \
\\(23\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\)) = \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"-", "1.9619793451909047`"}], 
   "\[InvisibleSpace]", "\<\"\[Degree]\"\>"}],
  SequenceForm[
  "The RG change of \!\(\*SubscriptBox[\(\[Theta]\), \(23\)]\) is :    \
\!\(\*SubscriptBox[\(\[Theta]\), \(23\)]\)(\!\(\*SubscriptBox[\(M\), \(GUT\)]\
\))-\!\(\*SubscriptBox[\(\[Theta]\), \(23\)]\)(\!\(\*SubscriptBox[\(M\), \
\(Z\)]\)) = ", -1.9619793451909047`, "\[Degree]"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847120664076*^9, 3.6368474637158937`*^9, 
  3.646456139612596*^9, 3.646475888647488*^9, 3.646477147918311*^9, 
  3.646477801170136*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Phases", "Subsection"],

Cell["Draw the plot for the phases", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"Plot", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"\[Delta]", "[", "t", "]"}], ",", 
       RowBox[{"\[CurlyPhi]1", "[", "t", "]"}], ",", 
       RowBox[{"\[CurlyPhi]2", "[", "t", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "tmin", ",", "tmax"}], "}"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"ImageSize", "\[Rule]", "400"}], ",", 
     RowBox[{"FrameLabel", "->", 
      RowBox[{"{", 
       RowBox[{
       "\"\<\!\(\*SubscriptBox[\(log\), \(10\)]\)(\[Mu]/GeV)\>\"", ",", 
        "\"\<\>\""}], "}"}]}], ",", 
     RowBox[{"FrameTicks", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"RGELogTicksLabeled", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"i", "*", "45"}], ",", 
            RowBox[{
             RowBox[{"ToString", "[", 
              RowBox[{"i", "*", "45"}], "]"}], "<>", "\"\<\[Degree]\>\""}]}], 
           "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", "0", ",", "8"}], "}"}]}], "]"}], ",", 
        RowBox[{"RGELogTicks", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"i", "*", "45"}], ",", "\"\<\>\""}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", "0", ",", "12"}], "}"}]}], "]"}]}], "}"}]}], ",",
      "\[IndentingNewLine]", 
     RowBox[{"PlotRange", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"All", ",", "All"}], "}"}]}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"Red", ",", "Blue", ",", "ForestGreen"}], "}"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"Prolog", "\[Rule]", 
      RowBox[{"{", "ShadowEFT", "}"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"Epilog", "\[Rule]", 
      RowBox[{"{", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"StyleForm", "[", 
           RowBox[{"\[Delta]", ",", 
            RowBox[{"FontColor", "\[Rule]", "Red"}], ",", 
            RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"tmin", ",", 
            RowBox[{"\[Delta]", "[", "tmin", "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "0.8"}], ",", "1.4"}], "}"}]}], "]"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"StyleForm", "[", 
           RowBox[{
            SubscriptBox["\[CurlyPhi]", "1"], ",", 
            RowBox[{"FontColor", "\[Rule]", "Blue"}], ",", 
            RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"tmin", ",", 
            RowBox[{"\[CurlyPhi]1", "[", "tmin", "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "0.8"}], ",", "1.4"}], "}"}]}], "]"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"StyleForm", "[", 
           RowBox[{
            SubscriptBox["\[CurlyPhi]", "2"], ",", 
            RowBox[{"FontColor", "\[Rule]", "ForestGreen"}], ",", 
            RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"tmin", ",", 
            RowBox[{"\[CurlyPhi]2", "[", "tmin", "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "0.8"}], ",", 
            RowBox[{"-", "1"}]}], "}"}]}], "]"}]}], "}"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"DisplayFunction", "\[Rule]", "Identity"}]}], 
    "\[IndentingNewLine]", "]"}], ",", 
   RowBox[{"DisplayFunction", "\[Rule]", "$DisplayFunction"}]}], 
  "]"}]], "Input",
 CellChangeTimes->{3.435655636439897*^9, 3.435655679288266*^9}],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {RGBColor[1, 0, 0], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwV03k41NsDx3HrzCBrlMpFEskX34as4dOE71hKiqlBEooYJUubnetS01VZ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     "]]}, 
   {RGBColor[0, 0, 1], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwl1Xs01dkXAHBJ91403JLS8Lskj/DFVyERdii6ipKSG7KShl9CDD08CkNN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     "]]}, 
   {RGBColor[0.1, 0.75, 0.2], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwV03s0lPkfwPEhZoZatCG/0hAily1FtMT3m4RGLtHWkluuI3IZl0pDN8ql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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{False, False},
  AxesLabel->{None, None},
  AxesOrigin->{2., 0},
  DisplayFunction->Identity,
  Epilog->{
    InsetBox[
     FormBox[
      TagBox[
       StyleBox["\[Delta]", FontColor -> RGBColor[1, 0, 0], FontSize -> 12], 
       StyleForm[#, FontColor -> RGBColor[1, 0, 0], FontSize -> 12]& ], 
      TraditionalForm], {1.9596137110710439`, 18.357868862591832`}, {-0.8, 
     1.4}], 
    InsetBox[
     FormBox[
      TagBox[
       StyleBox[
        SubscriptBox["\[CurlyPhi]", "1"], FontColor -> RGBColor[0, 0, 1], 
        FontSize -> 12], 
       StyleForm[#, FontColor -> RGBColor[0, 0, 1], FontSize -> 12]& ], 
      TraditionalForm], {1.9596137110710439`, 359.93380516812556`}, {-0.8, 
     1.4}], 
    InsetBox[
     FormBox[
      TagBox[
       StyleBox[
        SubscriptBox["\[CurlyPhi]", "2"], FontColor -> 
        RGBColor[0.1, 0.75, 0.2], FontSize -> 12], 
       StyleForm[#, FontColor -> RGBColor[0.1, 0.75, 0.2], FontSize -> 12]& ],
       TraditionalForm], {1.9596137110710439`, 
     0.022875536155592468`}, {-0.8, -1}]},
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{
     FormBox["\"\"", TraditionalForm], None}, {
     FormBox[
     "\"\\!\\(\\*SubscriptBox[\\(log\\), \\(10\\)]\\)(\[Mu]/GeV)\"", 
      TraditionalForm], None}},
  FrameTicks->{{{{0, 
       FormBox["\"0\[Degree]\"", TraditionalForm]}, {45, 
       FormBox["\"45\[Degree]\"", TraditionalForm]}, {90, 
       FormBox["\"90\[Degree]\"", TraditionalForm]}, {135, 
       FormBox["\"135\[Degree]\"", TraditionalForm]}, {180, 
       FormBox["\"180\[Degree]\"", TraditionalForm]}, {225, 
       FormBox["\"225\[Degree]\"", TraditionalForm]}, {270, 
       FormBox["\"270\[Degree]\"", TraditionalForm]}, {315, 
       FormBox["\"315\[Degree]\"", TraditionalForm]}, {360, 
       FormBox["\"360\[Degree]\"", TraditionalForm]}}, {{0, 
       FormBox["\"\"", TraditionalForm]}, {45, 
       FormBox["\"\"", TraditionalForm]}, {90, 
       FormBox["\"\"", TraditionalForm]}, {135, 
       FormBox["\"\"", TraditionalForm]}, {180, 
       FormBox["\"\"", TraditionalForm]}, {225, 
       FormBox["\"\"", TraditionalForm]}, {270, 
       FormBox["\"\"", TraditionalForm]}, {315, 
       FormBox["\"\"", TraditionalForm]}, {360, 
       FormBox["\"\"", TraditionalForm]}, {405, 
       FormBox["\"\"", TraditionalForm]}, {450, 
       FormBox["\"\"", TraditionalForm]}, {495, 
       FormBox["\"\"", TraditionalForm]}, {540, 
       FormBox["\"\"", TraditionalForm]}}}, {{{2, 
       FormBox["2", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["3", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["4", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["5", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["6", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["7", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["8", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["9", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["10", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["11", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["12", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["13", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["14", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["15", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["16", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["17", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["18", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["19", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}, {{2, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}}},
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImageSize->400,
  Method->{"DefaultBoundaryStyle" -> Automatic, "ScalingFunctions" -> None},
  PlotLabel->FormBox["\"\"", TraditionalForm],
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.05], 
     Scaled[0.05]}},
  Prolog->{{{
      GrayLevel[0.9], 
      RectangleBox[{10.561453473142157`, -10000}, {
       12.222494362608364`, 10000}]}, {
      GrayLevel[0.8], 
      RectangleBox[{12.222494362608364`, -10000}, {
       14.334981351900138`, 10000}]}, {
      GrayLevel[0.7], 
      RectangleBox[{14.334981351900138`, -10000}, {17, 10000}]}}},
  Ticks->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.636847121614245*^9, 3.636847464474209*^9, 
  3.646456141643083*^9, 3.646475889931695*^9, 3.646477149274476*^9, 
  3.646477802745575*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["RG change of the phases:", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<The RG change of \[Delta] is:   \[Delta](\!\(\*SubscriptBox[\(M\), \
\(GUT\)]\))-\[Delta](\!\(\*SubscriptBox[\(M\), \(Z\)]\)) = \>\"", ",", " ", 
    RowBox[{
     RowBox[{"\[Delta]", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "GUT"}], "]"}], "]"}], "-", 
     RowBox[{"\[Delta]", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "MZ"}], "]"}], "]"}]}], ",", "\"\<\[Degree]\>\""}], 
   "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<The RG change of \!\(\*SubscriptBox[\(\[CurlyPhi]\), \(1\)]\) is:   \
\!\(\*SubscriptBox[\(\[CurlyPhi]\), \(1\)]\)(\!\(\*SubscriptBox[\(M\), \
\(GUT\)]\))-\!\(\*SubscriptBox[\(\[CurlyPhi]\), \
\(1\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\)) = \>\"", ",", " ", 
    RowBox[{
     RowBox[{"\[CurlyPhi]1", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "GUT"}], "]"}], "]"}], "-", 
     RowBox[{"\[CurlyPhi]1", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "MZ"}], "]"}], "]"}]}], ",", "\"\<\[Degree]\>\""}], 
   "]"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<The RG change of \!\(\*SubscriptBox[\(\[CurlyPhi]\), \(2\)]\) is:   \
\!\(\*SubscriptBox[\(\[CurlyPhi]\), \(2\)]\)(\!\(\*SubscriptBox[\(M\), \
\(GUT\)]\))-\!\(\*SubscriptBox[\(\[CurlyPhi]\), \
\(2\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\)) = \>\"", ",", " ", 
    RowBox[{
     RowBox[{"\[CurlyPhi]2", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "GUT"}], "]"}], "]"}], "-", 
     RowBox[{"\[CurlyPhi]2", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "MZ"}], "]"}], "]"}]}], ",", "\"\<\[Degree]\>\""}], 
   "]"}], ";"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The RG change of \[Delta] is:   \
\[Delta](\\!\\(\\*SubscriptBox[\\(M\\), \
\\(GUT\\)]\\))-\[Delta](\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\)) = \"\>", 
   "\[InvisibleSpace]", "71.60560440448651`", 
   "\[InvisibleSpace]", "\<\"\[Degree]\"\>"}],
  SequenceForm[
  "The RG change of \[Delta] is:   \[Delta](\!\(\*SubscriptBox[\(M\), \
\(GUT\)]\))-\[Delta](\!\(\*SubscriptBox[\(M\), \(Z\)]\)) = ", 
   71.60560440448651, "\[Degree]"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847122923572*^9, 3.636847464526793*^9, 
  3.646456143029172*^9, 3.646475890002245*^9, 3.6464771493776493`*^9, 
  3.646477802884087*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The RG change of \\!\\(\\*SubscriptBox[\\(\[CurlyPhi]\\), \\(1\
\\)]\\) is:   \\!\\(\\*SubscriptBox[\\(\[CurlyPhi]\\), \
\\(1\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \
\\(GUT\\)]\\))-\\!\\(\\*SubscriptBox[\\(\[CurlyPhi]\\), \
\\(1\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\)) = \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"-", "0.019962031038630812`"}], 
   "\[InvisibleSpace]", "\<\"\[Degree]\"\>"}],
  SequenceForm[
  "The RG change of \!\(\*SubscriptBox[\(\[CurlyPhi]\), \(1\)]\) is:   \
\!\(\*SubscriptBox[\(\[CurlyPhi]\), \(1\)]\)(\!\(\*SubscriptBox[\(M\), \
\(GUT\)]\))-\!\(\*SubscriptBox[\(\[CurlyPhi]\), \
\(1\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\)) = ", -0.019962031038630812`, 
   "\[Degree]"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847122923572*^9, 3.636847464526793*^9, 
  3.646456143029172*^9, 3.646475890002245*^9, 3.6464771493776493`*^9, 
  3.646477802885468*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The RG change of \\!\\(\\*SubscriptBox[\\(\[CurlyPhi]\\), \\(2\
\\)]\\) is:   \\!\\(\\*SubscriptBox[\\(\[CurlyPhi]\\), \
\\(2\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \
\\(GUT\\)]\\))-\\!\\(\\*SubscriptBox[\\(\[CurlyPhi]\\), \
\\(2\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\)) = \"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"-", "0.003958419119325651`"}], 
   "\[InvisibleSpace]", "\<\"\[Degree]\"\>"}],
  SequenceForm[
  "The RG change of \!\(\*SubscriptBox[\(\[CurlyPhi]\), \(2\)]\) is:   \
\!\(\*SubscriptBox[\(\[CurlyPhi]\), \(2\)]\)(\!\(\*SubscriptBox[\(M\), \
\(GUT\)]\))-\!\(\*SubscriptBox[\(\[CurlyPhi]\), \
\(2\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\)) = ", -0.003958419119325651, 
   "\[Degree]"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847122923572*^9, 3.636847464526793*^9, 
  3.646456143029172*^9, 3.646475890002245*^9, 3.6464771493776493`*^9, 
  3.646477802886149*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Masses", "Subsection"],

Cell[CellGroupData[{

Cell["mass eigenvalues", "Subsubsection"],

Cell["Draw the plot for the mass eigenvalues", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"tlabel", "=", "4"}], ";"}], "   ", 
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{"Position", " ", "where", " ", "the", " ", "labels", " ", "m1"}], 
    ",", "m2", ",", 
    RowBox[{"m3", " ", "should", " ", "be", " ", "placed"}]}], " ", 
   "*)"}]}], "\[IndentingNewLine]", 
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"Plot", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"m1", "[", "t", "]"}], ",", 
       RowBox[{"m2", "[", "t", "]"}], ",", 
       RowBox[{"m3", "[", "t", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "tmin", ",", "tmax"}], "}"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"ImageSize", "\[Rule]", "400"}], ",", 
     RowBox[{"FrameLabel", "->", 
      RowBox[{"{", 
       RowBox[{
       "\"\<\!\(\*SubscriptBox[\(log\), \(10\)]\)(\[Mu]/GeV)\>\"", ",", 
        "\"\<\!\(\*SubscriptBox[\(m\), \(i\)]\) [eV]\>\""}], "}"}]}], ",", 
     RowBox[{"FrameTicks", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"RGELogTicksLabeled", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"i", "*", "0.01"}], ",", 
            RowBox[{"PaddedForm", "[", 
             RowBox[{
              RowBox[{"i", "*", "0.01"}], ",", 
              RowBox[{"{", 
               RowBox[{"3", ",", "2"}], "}"}]}], "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", "0", ",", "50"}], "}"}]}], "]"}], ",", 
        RowBox[{"RGELogTicks", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"i", "*", "0.01"}], ",", "\"\<\>\""}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", "0", ",", "50"}], "}"}]}], "]"}]}], "}"}]}], ",",
      "\[IndentingNewLine]", 
     RowBox[{"PlotRange", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"All", ",", "All"}], "}"}]}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"Red", ",", "Blue", ",", "ForestGreen"}], "}"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"Prolog", "\[Rule]", 
      RowBox[{"{", "ShadowEFT", "}"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"Epilog", "\[Rule]", 
      RowBox[{"{", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"StyleForm", "[", 
           RowBox[{
            SubscriptBox["m", "1"], ",", 
            RowBox[{"FontColor", "\[Rule]", "Red"}], ",", 
            RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"tlabel", ",", 
            RowBox[{"m1", "[", "tlabel", "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "0.8"}], ",", "1.4"}], "}"}]}], "]"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"StyleForm", "[", 
           RowBox[{
            SubscriptBox["m", "2"], ",", 
            RowBox[{"FontColor", "\[Rule]", "Blue"}], ",", 
            RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"tlabel", ",", 
            RowBox[{"m2", "[", "tlabel", "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "0.8"}], ",", 
            RowBox[{"-", "2"}]}], "}"}]}], "]"}], ",", "\[IndentingNewLine]", 
        
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"StyleForm", "[", 
           RowBox[{
            SubscriptBox["m", "3"], ",", 
            RowBox[{"FontColor", "\[Rule]", "ForestGreen"}], ",", 
            RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"tlabel", ",", 
            RowBox[{"m3", "[", "tlabel", "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "0.8"}], ",", "1.4"}], "}"}]}], "]"}]}], "}"}]}], 
     ",", "\[IndentingNewLine]", 
     RowBox[{"DisplayFunction", "\[Rule]", "Identity"}]}], 
    "\[IndentingNewLine]", "]"}], ",", 
   RowBox[{"DisplayFunction", "\[Rule]", "$DisplayFunction"}]}], 
  "]"}]}], "Input",
 CellChangeTimes->{3.435655642671777*^9, 3.435655682603495*^9}],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {RGBColor[1, 0, 0], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwV03k4lc0bB3DrOYeKI/u+ZjsPjoRSzG17hYqkhZCUKLuyJFtUKnU8nEcq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     "]]}, 
   {RGBColor[0, 0, 1], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwV0nk4lOsbwHHrGJxs2TrKHjKvMZYQcd8VKS1KOuEgUeikyDkiSaVkK2Xe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     "]]}, 
   {RGBColor[0.1, 0.75, 0.2], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwV03k41NsfB3CKGfsgSte+hPjKKFvC55NoRJeSbvakhULk3tRlKCRqSrK0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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{False, False},
  AxesLabel->{None, None},
  AxesOrigin->{2., 0.012},
  DisplayFunction->Identity,
  Epilog->{
    InsetBox[
     FormBox[
      TagBox[
       StyleBox[
        SubscriptBox["m", "1"], FontColor -> RGBColor[1, 0, 0], FontSize -> 
        12], StyleForm[#, FontColor -> RGBColor[1, 0, 0], FontSize -> 12]& ], 
      TraditionalForm], {4, 0.01661933208457743}, {-0.8, 1.4}], 
    InsetBox[
     FormBox[
      TagBox[
       StyleBox[
        SubscriptBox["m", "2"], FontColor -> RGBColor[0, 0, 1], FontSize -> 
        12], StyleForm[#, FontColor -> RGBColor[0, 0, 1], FontSize -> 12]& ], 
      TraditionalForm], {4, 0.01933301215581583}, {-0.8, -2}], 
    InsetBox[
     FormBox[
      TagBox[
       StyleBox[
        SubscriptBox["m", "3"], FontColor -> RGBColor[0.1, 0.75, 0.2], 
        FontSize -> 12], 
       StyleForm[#, FontColor -> RGBColor[0.1, 0.75, 0.2], FontSize -> 12]& ],
       TraditionalForm], {4, 0.05482118382023045}, {-0.8, 1.4}]},
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{
     FormBox[
     "\"\\!\\(\\*SubscriptBox[\\(m\\), \\(i\\)]\\) [eV]\"", TraditionalForm], 
     None}, {
     FormBox[
     "\"\\!\\(\\*SubscriptBox[\\(log\\), \\(10\\)]\\)(\[Mu]/GeV)\"", 
      TraditionalForm], None}},
  FrameTicks->{{{{0., 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.00\"", 0., AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.01, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.01\"", 0.01, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.02, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.02\"", 0.02, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.03, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.03\"", 0.03, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.04, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.04\"", 0.04, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.05, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.05\"", 0.05, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.06, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.06\"", 0.06, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.07, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.07\"", 0.07, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.08, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.08\"", 0.08, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.09, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.09\"", 0.09, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.1, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.10\"", 0.1, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.11, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.11\"", 0.11, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.12, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.12\"", 0.12, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.13, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.13\"", 0.13, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.14, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.14\"", 0.14, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.15, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.15\"", 0.15, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.16, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.16\"", 0.16, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.17, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.17\"", 0.17, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.18, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.18\"", 0.18, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.19, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.19\"", 0.19, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.2, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.20\"", 0.2, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.21, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.21\"", 0.21, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.22, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.22\"", 0.22, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.23, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.23\"", 0.23, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.24, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.24\"", 0.24, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.25, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.25\"", 0.25, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.26, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.26\"", 0.26, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.27, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.27\"", 0.27, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.28, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.28\"", 0.28, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.29, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.29\"", 0.29, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.3, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.30\"", 0.3, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.31, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.31\"", 0.31, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.32, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.32\"", 0.32, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.33, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.33\"", 0.33, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.34, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.34\"", 0.34, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.35000000000000003`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\" 0.35\"", 0.35000000000000003`, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.36, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.36\"", 0.36, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.37, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.37\"", 0.37, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.38, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.38\"", 0.38, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.39, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.39\"", 0.39, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.4, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.40\"", 0.4, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.41000000000000003`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\" 0.41\"", 0.41000000000000003`, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.42, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.42\"", 0.42, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.43, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.43\"", 0.43, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.44, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.44\"", 0.44, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.45, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.45\"", 0.45, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.46, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.46\"", 0.46, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.47000000000000003`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\" 0.47\"", 0.47000000000000003`, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.48, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.48\"", 0.48, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.49, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.49\"", 0.49, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}, {0.5, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.50\"", 0.5, AutoDelete -> True], 
         PaddedForm[#, {3, 2}]& ], TraditionalForm]}}, {{0., 
       FormBox["\"\"", TraditionalForm]}, {0.01, 
       FormBox["\"\"", TraditionalForm]}, {0.02, 
       FormBox["\"\"", TraditionalForm]}, {0.03, 
       FormBox["\"\"", TraditionalForm]}, {0.04, 
       FormBox["\"\"", TraditionalForm]}, {0.05, 
       FormBox["\"\"", TraditionalForm]}, {0.06, 
       FormBox["\"\"", TraditionalForm]}, {0.07, 
       FormBox["\"\"", TraditionalForm]}, {0.08, 
       FormBox["\"\"", TraditionalForm]}, {0.09, 
       FormBox["\"\"", TraditionalForm]}, {0.1, 
       FormBox["\"\"", TraditionalForm]}, {0.11, 
       FormBox["\"\"", TraditionalForm]}, {0.12, 
       FormBox["\"\"", TraditionalForm]}, {0.13, 
       FormBox["\"\"", TraditionalForm]}, {0.14, 
       FormBox["\"\"", TraditionalForm]}, {0.15, 
       FormBox["\"\"", TraditionalForm]}, {0.16, 
       FormBox["\"\"", TraditionalForm]}, {0.17, 
       FormBox["\"\"", TraditionalForm]}, {0.18, 
       FormBox["\"\"", TraditionalForm]}, {0.19, 
       FormBox["\"\"", TraditionalForm]}, {0.2, 
       FormBox["\"\"", TraditionalForm]}, {0.21, 
       FormBox["\"\"", TraditionalForm]}, {0.22, 
       FormBox["\"\"", TraditionalForm]}, {0.23, 
       FormBox["\"\"", TraditionalForm]}, {0.24, 
       FormBox["\"\"", TraditionalForm]}, {0.25, 
       FormBox["\"\"", TraditionalForm]}, {0.26, 
       FormBox["\"\"", TraditionalForm]}, {0.27, 
       FormBox["\"\"", TraditionalForm]}, {0.28, 
       FormBox["\"\"", TraditionalForm]}, {0.29, 
       FormBox["\"\"", TraditionalForm]}, {0.3, 
       FormBox["\"\"", TraditionalForm]}, {0.31, 
       FormBox["\"\"", TraditionalForm]}, {0.32, 
       FormBox["\"\"", TraditionalForm]}, {0.33, 
       FormBox["\"\"", TraditionalForm]}, {0.34, 
       FormBox["\"\"", TraditionalForm]}, {0.35000000000000003`, 
       FormBox["\"\"", TraditionalForm]}, {0.36, 
       FormBox["\"\"", TraditionalForm]}, {0.37, 
       FormBox["\"\"", TraditionalForm]}, {0.38, 
       FormBox["\"\"", TraditionalForm]}, {0.39, 
       FormBox["\"\"", TraditionalForm]}, {0.4, 
       FormBox["\"\"", TraditionalForm]}, {0.41000000000000003`, 
       FormBox["\"\"", TraditionalForm]}, {0.42, 
       FormBox["\"\"", TraditionalForm]}, {0.43, 
       FormBox["\"\"", TraditionalForm]}, {0.44, 
       FormBox["\"\"", TraditionalForm]}, {0.45, 
       FormBox["\"\"", TraditionalForm]}, {0.46, 
       FormBox["\"\"", TraditionalForm]}, {0.47000000000000003`, 
       FormBox["\"\"", TraditionalForm]}, {0.48, 
       FormBox["\"\"", TraditionalForm]}, {0.49, 
       FormBox["\"\"", TraditionalForm]}, {0.5, 
       FormBox["\"\"", TraditionalForm]}}}, {{{2, 
       FormBox["2", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["3", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["4", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["5", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["6", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["7", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["8", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["9", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["10", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["11", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["12", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["13", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["14", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["15", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["16", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["17", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["18", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["19", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}, {{2, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}}},
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImageSize->400,
  Method->{"DefaultBoundaryStyle" -> Automatic, "ScalingFunctions" -> None},
  PlotLabel->FormBox["\"\"", TraditionalForm],
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.05], 
     Scaled[0.05]}},
  Prolog->{{{
      GrayLevel[0.9], 
      RectangleBox[{10.561453473142157`, -10000}, {
       12.222494362608364`, 10000}]}, {
      GrayLevel[0.8], 
      RectangleBox[{12.222494362608364`, -10000}, {
       14.334981351900138`, 10000}]}, {
      GrayLevel[0.7], 
      RectangleBox[{14.334981351900138`, -10000}, {17, 10000}]}}},
  Ticks->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.636847123706056*^9, 3.636847465241275*^9, 
  3.646456145462736*^9, 3.646475890883382*^9, 3.646477150678502*^9, 
  3.6464778042280807`*^9}]
}, Open  ]],

Cell["RG change of the masses:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<The RG change of \!\(\*SubscriptBox[\(m\), \(1\)]\) is:   \
\!\(\*SubscriptBox[\(m\), \(1\)]\)(\!\(\*SubscriptBox[\(M\), \(GUT\)]\))-\!\(\
\*SubscriptBox[\(m\), \(1\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\)) = \>\"", 
    ",", 
    RowBox[{
     RowBox[{"m1", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "GUT"}], "]"}], "]"}], "-", 
     RowBox[{"m1", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "MZ"}], "]"}], "]"}]}], ",", "\"\< eV\>\""}], "]"}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<The RG change of \!\(\*SubscriptBox[\(m\), \(2\)]\) is:   \
\!\(\*SubscriptBox[\(m\), \(2\)]\)(\!\(\*SubscriptBox[\(M\), \(GUT\)]\))-\!\(\
\*SubscriptBox[\(m\), \(2\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\)) = \>\"", 
    ",", 
    RowBox[{
     RowBox[{"m2", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "GUT"}], "]"}], "]"}], "-", 
     RowBox[{"m2", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "MZ"}], "]"}], "]"}]}], ",", "\"\< eV\>\""}], "]"}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<The RG change of \!\(\*SubscriptBox[\(m\), \(3\)]\) is:   \
\!\(\*SubscriptBox[\(m\), \(3\)]\)(\!\(\*SubscriptBox[\(M\), \(GUT\)]\))-\!\(\
\*SubscriptBox[\(m\), \(3\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\)) = \>\"", 
    ",", 
    RowBox[{
     RowBox[{"m3", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "GUT"}], "]"}], "]"}], "-", 
     RowBox[{"m3", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "MZ"}], "]"}], "]"}]}], ",", "\"\< eV\>\""}], "]"}],
   ";"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The RG change of \\!\\(\\*SubscriptBox[\\(m\\), \\(1\\)]\\) \
is:   \\!\\(\\*SubscriptBox[\\(m\\), \
\\(1\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \
\\(GUT\\)]\\))-\\!\\(\\*SubscriptBox[\\(m\\), \
\\(1\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\)) = \"\>", 
   "\[InvisibleSpace]", "0.005156542987091039`", 
   "\[InvisibleSpace]", "\<\" eV\"\>"}],
  SequenceForm[
  "The RG change of \!\(\*SubscriptBox[\(m\), \(1\)]\) is:   \
\!\(\*SubscriptBox[\(m\), \(1\)]\)(\!\(\*SubscriptBox[\(M\), \(GUT\)]\))-\!\(\
\*SubscriptBox[\(m\), \(1\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\)) = ", 
   0.005156542987091039, " eV"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847123754135*^9, 3.636847465317541*^9, 
  3.646456145752508*^9, 3.646475891169506*^9, 3.646477151080858*^9, 
  3.6464778043766336`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The RG change of \\!\\(\\*SubscriptBox[\\(m\\), \\(2\\)]\\) \
is:   \\!\\(\\*SubscriptBox[\\(m\\), \
\\(2\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \
\\(GUT\\)]\\))-\\!\\(\\*SubscriptBox[\\(m\\), \
\\(2\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\)) = \"\>", 
   "\[InvisibleSpace]", "0.006415547841794732`", 
   "\[InvisibleSpace]", "\<\" eV\"\>"}],
  SequenceForm[
  "The RG change of \!\(\*SubscriptBox[\(m\), \(2\)]\) is:   \
\!\(\*SubscriptBox[\(m\), \(2\)]\)(\!\(\*SubscriptBox[\(M\), \(GUT\)]\))-\!\(\
\*SubscriptBox[\(m\), \(2\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\)) = ", 
   0.006415547841794732, " eV"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847123754135*^9, 3.636847465317541*^9, 
  3.646456145752508*^9, 3.646475891169506*^9, 3.646477151080858*^9, 
  3.6464778043776703`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The RG change of \\!\\(\\*SubscriptBox[\\(m\\), \\(3\\)]\\) \
is:   \\!\\(\\*SubscriptBox[\\(m\\), \
\\(3\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \
\\(GUT\\)]\\))-\\!\\(\\*SubscriptBox[\\(m\\), \
\\(3\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\)) = \"\>", 
   "\[InvisibleSpace]", "0.01862017509591489`", 
   "\[InvisibleSpace]", "\<\" eV\"\>"}],
  SequenceForm[
  "The RG change of \!\(\*SubscriptBox[\(m\), \(3\)]\) is:   \
\!\(\*SubscriptBox[\(m\), \(3\)]\)(\!\(\*SubscriptBox[\(M\), \(GUT\)]\))-\!\(\
\*SubscriptBox[\(m\), \(3\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\)) = ", 
   0.01862017509591489, " eV"],
  Editable->False]], "Print",
 CellChangeTimes->{3.636847123754135*^9, 3.636847465317541*^9, 
  3.646456145752508*^9, 3.646475891169506*^9, 3.646477151080858*^9, 
  3.646477804378641*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["mass squared differences", "Subsubsection"],

Cell["\<\
Draw the plot for the mass squared differences (on a logarithmic scale)\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"Plot", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"Log", "[", 
        RowBox[{"10", ",", 
         RowBox[{"Abs", "[", 
          RowBox[{"\[CapitalDelta]m2sol", "[", "t", "]"}], "]"}]}], "]"}], 
       ",", 
       RowBox[{"Log", "[", 
        RowBox[{"10", ",", 
         RowBox[{"Abs", "[", 
          RowBox[{"\[CapitalDelta]m2atm", "[", "t", "]"}], "]"}]}], "]"}]}], 
      "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "tmin", ",", "tmax"}], "}"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"ImageSize", "\[Rule]", "400"}], ",", 
     RowBox[{"FrameLabel", "->", 
      RowBox[{"{", 
       RowBox[{
       "\"\<\!\(\*SubscriptBox[\(log\), \(10\)]\)(\[Mu]/GeV)\>\"", ",", 
        "\"\<\!\(\*SubsuperscriptBox[\(\[CapitalDelta]m\), \(i\), \(2\)]\) \
[\!\(\*SuperscriptBox[\(eV\), \(2\)]\)]\>\""}], "}"}]}], ",", 
     RowBox[{"FrameTicks", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"RGELogTicksLabeled", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", 
        RowBox[{"RGELogTicksLabeledNegExp", "[", 
         RowBox[{
          RowBox[{"-", "6"}], ",", 
          RowBox[{"-", "1"}]}], "]"}], ",", 
        RowBox[{"RGELogTicks", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", 
        RowBox[{"RGELogTicks", "[", 
         RowBox[{
          RowBox[{"-", "6"}], ",", 
          RowBox[{"-", "1"}]}], "]"}]}], "}"}]}], ",", "\[IndentingNewLine]", 
     
     RowBox[{"PlotRange", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"All", ",", "All"}], "}"}]}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"ForestGreen", ",", "Blue"}], "}"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"Prolog", "\[Rule]", 
      RowBox[{"{", "ShadowEFT", "}"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"Epilog", "\[Rule]", 
      RowBox[{"{", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"StyleForm", "[", 
           RowBox[{
            SubsuperscriptBox["\[CapitalDelta]m", "sol", "2"], ",", 
            RowBox[{"FontColor", "\[Rule]", "ForestGreen"}], ",", 
            RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"tmin", ",", 
            RowBox[{"Log", "[", 
             RowBox[{"10", ",", 
              RowBox[{"Abs", "[", 
               RowBox[{"\[CapitalDelta]m2sol", "[", "tmin", "]"}], "]"}]}], 
             "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "0.8"}], ",", 
            RowBox[{"-", "2"}]}], "}"}]}], "]"}], ",", "\[IndentingNewLine]", 
        
        RowBox[{"Text", "[", 
         RowBox[{
          RowBox[{"StyleForm", "[", 
           RowBox[{
            SubsuperscriptBox["\[CapitalDelta]m", "atm", "2"], ",", 
            RowBox[{"FontColor", "\[Rule]", "Blue"}], ",", 
            RowBox[{"FontSize", "\[Rule]", "12"}]}], "]"}], ",", 
          RowBox[{"{", 
           RowBox[{"tmin", ",", 
            RowBox[{"Log", "[", 
             RowBox[{"10", ",", 
              RowBox[{"Abs", "[", 
               RowBox[{"\[CapitalDelta]m2atm", "[", "tmin", "]"}], "]"}]}], 
             "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"-", "0.8"}], ",", "1.4"}], "}"}]}], "]"}]}], "}"}]}], 
     ",", "\[IndentingNewLine]", 
     RowBox[{"DisplayFunction", "\[Rule]", "Identity"}]}], 
    "\[IndentingNewLine]", "]"}], ",", 
   RowBox[{"DisplayFunction", "\[Rule]", "$DisplayFunction"}]}], 
  "]"}]], "Input",
 CellChangeTimes->{3.435655651153079*^9, 3.4361001025721*^9}],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {RGBColor[0.1, 0.75, 0.2], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwV1Hk41dkfB3DEvRfje7+2EiFL1i+uvaH4jOhapoQiQoiQXUNlrVGpIZW1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     "]]}, 
   {RGBColor[0, 0, 1], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwV1Hk0ldsbB3C873vOkXKoFBmTUF68MmZ8KjrhRqkbR5IrQmZFXaGBbmUq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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{False, False},
  AxesLabel->{None, None},
  AxesOrigin->{2., -4.2},
  DisplayFunction->Identity,
  Epilog->{
    InsetBox[
     FormBox[
      TagBox[
       StyleBox[
        SubsuperscriptBox["\[CapitalDelta]m", "sol", "2"], FontColor -> 
        RGBColor[0.1, 0.75, 0.2], FontSize -> 12], 
       StyleForm[#, FontColor -> RGBColor[0.1, 0.75, 0.2], FontSize -> 12]& ],
       TraditionalForm], {
     1.9596137110710439`, -4.112358170735094}, {-0.8, -2}], 
    InsetBox[
     FormBox[
      TagBox[
       StyleBox[
        SubsuperscriptBox["\[CapitalDelta]m", "atm", "2"], FontColor -> 
        RGBColor[0, 0, 1], FontSize -> 12], 
       StyleForm[#, FontColor -> RGBColor[0, 0, 1], FontSize -> 12]& ], 
      TraditionalForm], {1.9596137110710439`, -2.679231641338273}, {-0.8, 
     1.4}]},
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{
     FormBox[
     "\"\\!\\(\\*SubsuperscriptBox[\\(\[CapitalDelta]m\\), \\(i\\), \
\\(2\\)]\\) [\\!\\(\\*SuperscriptBox[\\(eV\\), \\(2\\)]\\)]\"", 
      TraditionalForm], None}, {
     FormBox[
     "\"\\!\\(\\*SubscriptBox[\\(log\\), \\(10\\)]\\)(\[Mu]/GeV)\"", 
      TraditionalForm], None}},
  FrameTicks->{{{{-6, 
       FormBox["\"\\!\\(10\\^\\(-6\\)\\)\"", TraditionalForm]}, {
       NCache[-Log[10]^(-1) Log[500000], -5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[1000000, 3]], -5.5228787452803365`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[250000], -5.397940008672037], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[200000], -5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[500000, 3]], -5.221848749616355], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[1000000, 7]], -5.154901959985742], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[125000], -5.096910013008055], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[1000000, 9]], -5.045757490560675], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {-5, 
       FormBox["\"\\!\\(10\\^\\(-5\\)\\)\"", TraditionalForm]}, {
       NCache[-Log[10]^(-1) Log[50000], -4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[100000, 3]], -4.522878745280337], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[25000], -4.397940008672037], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[20000], -4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[50000, 3]], -4.221848749616356], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[100000, 7]], -4.154901959985742], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[12500], -4.096910013008055], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[100000, 9]], -4.045757490560674], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {-4, 
       FormBox["\"\\!\\(10\\^\\(-4\\)\\)\"", TraditionalForm]}, {
       NCache[-Log[10]^(-1) Log[5000], -3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[10000, 3]], -3.522878745280337], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[2500], -3.397940008672037], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[2000], -3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[5000, 3]], -3.221848749616356], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[10000, 7]], -3.1549019599857426`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[1250], -3.096910013008056], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[10000, 9]], -3.045757490560675], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {-3, 
       FormBox["\"\\!\\(10\\^\\(-3\\)\\)\"", TraditionalForm]}, {
       NCache[-Log[10]^(-1) Log[500], -2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[1000, 3]], -2.5228787452803374`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[250], -2.397940008672037], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[200], -2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[500, 3]], -2.221848749616356], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[1000, 7]], -2.1549019599857426`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[125], -2.0969100130080562`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[1000, 9]], -2.045757490560675], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {-2, 
       FormBox["\"\\!\\(10\\^\\(-2\\)\\)\"", TraditionalForm]}, {
       NCache[-Log[10]^(-1) Log[50], -1.6989700043360185`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[100, 3]], -1.5228787452803374`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[25], -1.3979400086720373`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[20], -1.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[50, 3]], -1.2218487496163561`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[100, 7]], -1.1549019599857429`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[25, 2]], -1.0969100130080562`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[100, 9]], -1.0457574905606748`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {-1, 
       FormBox["\"\\!\\(10\\^\\(-1\\)\\)\"", TraditionalForm]}, {
       NCache[-Log[5]/Log[10], -0.6989700043360186], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[
          Rational[10, 3]]/Log[10], -0.5228787452803375], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[
          Rational[5, 2]]/Log[10], -0.39794000867203755`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[2]/Log[10], -0.30102999566398114`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[
          Rational[5, 3]]/Log[10], -0.22184874961635634`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[
          Rational[10, 7]]/Log[10], -0.15490195998574316`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[
          Rational[5, 4]]/Log[10], -0.0969100130080564], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[
          Rational[10, 9]]/Log[10], -0.04575749056067514], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}, {{-6, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[-Log[10]^(-1) Log[500000], -5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[1000000, 3]], -5.5228787452803365`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[250000], -5.397940008672037], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[200000], -5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[500000, 3]], -5.221848749616355], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[1000000, 7]], -5.154901959985742], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[125000], -5.096910013008055], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[1000000, 9]], -5.045757490560675], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {-5, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[-Log[10]^(-1) Log[50000], -4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[100000, 3]], -4.522878745280337], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[25000], -4.397940008672037], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[20000], -4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[50000, 3]], -4.221848749616356], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[100000, 7]], -4.154901959985742], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[12500], -4.096910013008055], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[100000, 9]], -4.045757490560674], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {-4, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[-Log[10]^(-1) Log[5000], -3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[10000, 3]], -3.522878745280337], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[2500], -3.397940008672037], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[2000], -3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[5000, 3]], -3.221848749616356], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[10000, 7]], -3.1549019599857426`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[1250], -3.096910013008056], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[10000, 9]], -3.045757490560675], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {-3, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[-Log[10]^(-1) Log[500], -2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[1000, 3]], -2.5228787452803374`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[250], -2.397940008672037], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[200], -2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[500, 3]], -2.221848749616356], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[1000, 7]], -2.1549019599857426`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[125], -2.0969100130080562`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[1000, 9]], -2.045757490560675], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {-2, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[-Log[10]^(-1) Log[50], -1.6989700043360185`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[100, 3]], -1.5228787452803374`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[25], -1.3979400086720373`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[20], -1.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[50, 3]], -1.2218487496163561`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[100, 7]], -1.1549019599857429`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[25, 2]], -1.0969100130080562`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[10]^(-1) Log[
          Rational[100, 9]], -1.0457574905606748`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {-1, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[-Log[5]/Log[10], -0.6989700043360186], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[
          Rational[10, 3]]/Log[10], -0.5228787452803375], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[
          Rational[5, 2]]/Log[10], -0.39794000867203755`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[2]/Log[10], -0.30102999566398114`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[
          Rational[5, 3]]/Log[10], -0.22184874961635634`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[
          Rational[10, 7]]/Log[10], -0.15490195998574316`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[
          Rational[5, 4]]/Log[10], -0.0969100130080564], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[-Log[
          Rational[10, 9]]/Log[10], -0.04575749056067514], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}}, {{{2, 
       FormBox["2", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["3", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["4", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["5", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["6", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["7", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["8", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["9", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["10", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["11", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["12", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["13", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["14", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["15", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["16", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["17", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["18", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["19", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}, {{2, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}}},
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImageSize->400,
  Method->{"DefaultBoundaryStyle" -> Automatic, "ScalingFunctions" -> None},
  PlotLabel->FormBox["\"\"", TraditionalForm],
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.05], 
     Scaled[0.05]}},
  Prolog->{{{
      GrayLevel[0.9], 
      RectangleBox[{10.561453473142157`, -10000}, {
       12.222494362608364`, 10000}]}, {
      GrayLevel[0.8], 
      RectangleBox[{12.222494362608364`, -10000}, {
       14.334981351900138`, 10000}]}, {
      GrayLevel[0.7], 
      RectangleBox[{14.334981351900138`, -10000}, {17, 10000}]}}},
  Ticks->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.636847124652256*^9, 3.6368474661977262`*^9, 
  3.646456146963194*^9, 3.6464758921807203`*^9, 3.64647715317387*^9, 
  3.6464778057300053`*^9}]
}, Open  ]],

Cell["\<\
Draw the plot for the solar mass squared difference (on a linear scale)\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"Plot", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"\[CapitalDelta]m2sol", "[", "t", "]"}], "*", 
      RowBox[{"10", "^", "5"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "tmin", ",", "tmax"}], "}"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"ImageSize", "\[Rule]", "400"}], ",", 
     RowBox[{"FrameLabel", "->", 
      RowBox[{"{", 
       RowBox[{
       "\"\<\!\(\*SubscriptBox[\(log\), \(10\)]\)(\[Mu]/GeV)\>\"", ",", 
        "\"\<\!\(\*SubsuperscriptBox[\(\[CapitalDelta]m\), \(sol\), \(2\)]\) \
[\!\(\*SuperscriptBox[\(10\), \
\(-5\)]\)\[ThinSpace]\!\(\*SuperscriptBox[\(eV\), \(2\)]\)]\>\""}], "}"}]}], 
     ",", 
     RowBox[{"FrameTicks", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"RGELogTicksLabeled", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"i", ",", "i"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", 
            RowBox[{"-", "100"}], ",", "100"}], "}"}]}], "]"}], ",", 
        RowBox[{"RGELogTicks", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{"i", ",", "\"\<\>\""}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", 
            RowBox[{"-", "100"}], ",", "100"}], "}"}]}], "]"}]}], "}"}]}], 
     ",", "\[IndentingNewLine]", 
     RowBox[{"PlotRange", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"All", ",", "All"}], "}"}]}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"{", "ForestGreen", "}"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"Prolog", "\[Rule]", 
      RowBox[{"{", "ShadowEFT", "}"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"DisplayFunction", "\[Rule]", "Identity"}]}], 
    "\[IndentingNewLine]", "]"}], ",", 
   RowBox[{"DisplayFunction", "\[Rule]", "$DisplayFunction"}]}], 
  "]"}]], "Input",
 CellChangeTimes->{3.435655655075964*^9, 3.435655686756006*^9}],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {RGBColor[0.1, 0.75, 0.2], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwVkns01Pkbx8kYI5tLLtmtHTNfd/P9zoxcKsK7VsaSa7U1SDZE26Sa0hVd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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{False, False},
  AxesLabel->{None, None},
  AxesOrigin->{2., 7.5},
  DisplayFunction->Identity,
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{
     FormBox[
     "\"\\!\\(\\*SubsuperscriptBox[\\(\[CapitalDelta]m\\), \\(sol\\), \
\\(2\\)]\\) [\\!\\(\\*SuperscriptBox[\\(10\\), \\(-5\\)]\\)\[ThinSpace]\\!\\(\
\\*SuperscriptBox[\\(eV\\), \\(2\\)]\\)]\"", TraditionalForm], None}, {
     FormBox[
     "\"\\!\\(\\*SubscriptBox[\\(log\\), \\(10\\)]\\)(\[Mu]/GeV)\"", 
      TraditionalForm], None}},
  FrameTicks->{{{{-100, 
       FormBox[
        RowBox[{"-", "100"}], TraditionalForm]}, {-99, 
       FormBox[
        RowBox[{"-", "99"}], TraditionalForm]}, {-98, 
       FormBox[
        RowBox[{"-", "98"}], TraditionalForm]}, {-97, 
       FormBox[
        RowBox[{"-", "97"}], TraditionalForm]}, {-96, 
       FormBox[
        RowBox[{"-", "96"}], TraditionalForm]}, {-95, 
       FormBox[
        RowBox[{"-", "95"}], TraditionalForm]}, {-94, 
       FormBox[
        RowBox[{"-", "94"}], TraditionalForm]}, {-93, 
       FormBox[
        RowBox[{"-", "93"}], TraditionalForm]}, {-92, 
       FormBox[
        RowBox[{"-", "92"}], TraditionalForm]}, {-91, 
       FormBox[
        RowBox[{"-", "91"}], TraditionalForm]}, {-90, 
       FormBox[
        RowBox[{"-", "90"}], TraditionalForm]}, {-89, 
       FormBox[
        RowBox[{"-", "89"}], TraditionalForm]}, {-88, 
       FormBox[
        RowBox[{"-", "88"}], TraditionalForm]}, {-87, 
       FormBox[
        RowBox[{"-", "87"}], TraditionalForm]}, {-86, 
       FormBox[
        RowBox[{"-", "86"}], TraditionalForm]}, {-85, 
       FormBox[
        RowBox[{"-", "85"}], TraditionalForm]}, {-84, 
       FormBox[
        RowBox[{"-", "84"}], TraditionalForm]}, {-83, 
       FormBox[
        RowBox[{"-", "83"}], TraditionalForm]}, {-82, 
       FormBox[
        RowBox[{"-", "82"}], TraditionalForm]}, {-81, 
       FormBox[
        RowBox[{"-", "81"}], TraditionalForm]}, {-80, 
       FormBox[
        RowBox[{"-", "80"}], TraditionalForm]}, {-79, 
       FormBox[
        RowBox[{"-", "79"}], TraditionalForm]}, {-78, 
       FormBox[
        RowBox[{"-", "78"}], TraditionalForm]}, {-77, 
       FormBox[
        RowBox[{"-", "77"}], TraditionalForm]}, {-76, 
       FormBox[
        RowBox[{"-", "76"}], TraditionalForm]}, {-75, 
       FormBox[
        RowBox[{"-", "75"}], TraditionalForm]}, {-74, 
       FormBox[
        RowBox[{"-", "74"}], TraditionalForm]}, {-73, 
       FormBox[
        RowBox[{"-", "73"}], TraditionalForm]}, {-72, 
       FormBox[
        RowBox[{"-", "72"}], TraditionalForm]}, {-71, 
       FormBox[
        RowBox[{"-", "71"}], TraditionalForm]}, {-70, 
       FormBox[
        RowBox[{"-", "70"}], TraditionalForm]}, {-69, 
       FormBox[
        RowBox[{"-", "69"}], TraditionalForm]}, {-68, 
       FormBox[
        RowBox[{"-", "68"}], TraditionalForm]}, {-67, 
       FormBox[
        RowBox[{"-", "67"}], TraditionalForm]}, {-66, 
       FormBox[
        RowBox[{"-", "66"}], TraditionalForm]}, {-65, 
       FormBox[
        RowBox[{"-", "65"}], TraditionalForm]}, {-64, 
       FormBox[
        RowBox[{"-", "64"}], TraditionalForm]}, {-63, 
       FormBox[
        RowBox[{"-", "63"}], TraditionalForm]}, {-62, 
       FormBox[
        RowBox[{"-", "62"}], TraditionalForm]}, {-61, 
       FormBox[
        RowBox[{"-", "61"}], TraditionalForm]}, {-60, 
       FormBox[
        RowBox[{"-", "60"}], TraditionalForm]}, {-59, 
       FormBox[
        RowBox[{"-", "59"}], TraditionalForm]}, {-58, 
       FormBox[
        RowBox[{"-", "58"}], TraditionalForm]}, {-57, 
       FormBox[
        RowBox[{"-", "57"}], TraditionalForm]}, {-56, 
       FormBox[
        RowBox[{"-", "56"}], TraditionalForm]}, {-55, 
       FormBox[
        RowBox[{"-", "55"}], TraditionalForm]}, {-54, 
       FormBox[
        RowBox[{"-", "54"}], TraditionalForm]}, {-53, 
       FormBox[
        RowBox[{"-", "53"}], TraditionalForm]}, {-52, 
       FormBox[
        RowBox[{"-", "52"}], TraditionalForm]}, {-51, 
       FormBox[
        RowBox[{"-", "51"}], TraditionalForm]}, {-50, 
       FormBox[
        RowBox[{"-", "50"}], TraditionalForm]}, {-49, 
       FormBox[
        RowBox[{"-", "49"}], TraditionalForm]}, {-48, 
       FormBox[
        RowBox[{"-", "48"}], TraditionalForm]}, {-47, 
       FormBox[
        RowBox[{"-", "47"}], TraditionalForm]}, {-46, 
       FormBox[
        RowBox[{"-", "46"}], TraditionalForm]}, {-45, 
       FormBox[
        RowBox[{"-", "45"}], TraditionalForm]}, {-44, 
       FormBox[
        RowBox[{"-", "44"}], TraditionalForm]}, {-43, 
       FormBox[
        RowBox[{"-", "43"}], TraditionalForm]}, {-42, 
       FormBox[
        RowBox[{"-", "42"}], TraditionalForm]}, {-41, 
       FormBox[
        RowBox[{"-", "41"}], TraditionalForm]}, {-40, 
       FormBox[
        RowBox[{"-", "40"}], TraditionalForm]}, {-39, 
       FormBox[
        RowBox[{"-", "39"}], TraditionalForm]}, {-38, 
       FormBox[
        RowBox[{"-", "38"}], TraditionalForm]}, {-37, 
       FormBox[
        RowBox[{"-", "37"}], TraditionalForm]}, {-36, 
       FormBox[
        RowBox[{"-", "36"}], TraditionalForm]}, {-35, 
       FormBox[
        RowBox[{"-", "35"}], TraditionalForm]}, {-34, 
       FormBox[
        RowBox[{"-", "34"}], TraditionalForm]}, {-33, 
       FormBox[
        RowBox[{"-", "33"}], TraditionalForm]}, {-32, 
       FormBox[
        RowBox[{"-", "32"}], TraditionalForm]}, {-31, 
       FormBox[
        RowBox[{"-", "31"}], TraditionalForm]}, {-30, 
       FormBox[
        RowBox[{"-", "30"}], TraditionalForm]}, {-29, 
       FormBox[
        RowBox[{"-", "29"}], TraditionalForm]}, {-28, 
       FormBox[
        RowBox[{"-", "28"}], TraditionalForm]}, {-27, 
       FormBox[
        RowBox[{"-", "27"}], TraditionalForm]}, {-26, 
       FormBox[
        RowBox[{"-", "26"}], TraditionalForm]}, {-25, 
       FormBox[
        RowBox[{"-", "25"}], TraditionalForm]}, {-24, 
       FormBox[
        RowBox[{"-", "24"}], TraditionalForm]}, {-23, 
       FormBox[
        RowBox[{"-", "23"}], TraditionalForm]}, {-22, 
       FormBox[
        RowBox[{"-", "22"}], TraditionalForm]}, {-21, 
       FormBox[
        RowBox[{"-", "21"}], TraditionalForm]}, {-20, 
       FormBox[
        RowBox[{"-", "20"}], TraditionalForm]}, {-19, 
       FormBox[
        RowBox[{"-", "19"}], TraditionalForm]}, {-18, 
       FormBox[
        RowBox[{"-", "18"}], TraditionalForm]}, {-17, 
       FormBox[
        RowBox[{"-", "17"}], TraditionalForm]}, {-16, 
       FormBox[
        RowBox[{"-", "16"}], TraditionalForm]}, {-15, 
       FormBox[
        RowBox[{"-", "15"}], TraditionalForm]}, {-14, 
       FormBox[
        RowBox[{"-", "14"}], TraditionalForm]}, {-13, 
       FormBox[
        RowBox[{"-", "13"}], TraditionalForm]}, {-12, 
       FormBox[
        RowBox[{"-", "12"}], TraditionalForm]}, {-11, 
       FormBox[
        RowBox[{"-", "11"}], TraditionalForm]}, {-10, 
       FormBox[
        RowBox[{"-", "10"}], TraditionalForm]}, {-9, 
       FormBox[
        RowBox[{"-", "9"}], TraditionalForm]}, {-8, 
       FormBox[
        RowBox[{"-", "8"}], TraditionalForm]}, {-7, 
       FormBox[
        RowBox[{"-", "7"}], TraditionalForm]}, {-6, 
       FormBox[
        RowBox[{"-", "6"}], TraditionalForm]}, {-5, 
       FormBox[
        RowBox[{"-", "5"}], TraditionalForm]}, {-4, 
       FormBox[
        RowBox[{"-", "4"}], TraditionalForm]}, {-3, 
       FormBox[
        RowBox[{"-", "3"}], TraditionalForm]}, {-2, 
       FormBox[
        RowBox[{"-", "2"}], TraditionalForm]}, {-1, 
       FormBox[
        RowBox[{"-", "1"}], TraditionalForm]}, {0, 
       FormBox["0", TraditionalForm]}, {1, 
       FormBox["1", TraditionalForm]}, {2, 
       FormBox["2", TraditionalForm]}, {3, 
       FormBox["3", TraditionalForm]}, {4, 
       FormBox["4", TraditionalForm]}, {5, 
       FormBox["5", TraditionalForm]}, {6, 
       FormBox["6", TraditionalForm]}, {7, 
       FormBox["7", TraditionalForm]}, {8, 
       FormBox["8", TraditionalForm]}, {9, 
       FormBox["9", TraditionalForm]}, {10, 
       FormBox["10", TraditionalForm]}, {11, 
       FormBox["11", TraditionalForm]}, {12, 
       FormBox["12", TraditionalForm]}, {13, 
       FormBox["13", TraditionalForm]}, {14, 
       FormBox["14", TraditionalForm]}, {15, 
       FormBox["15", TraditionalForm]}, {16, 
       FormBox["16", TraditionalForm]}, {17, 
       FormBox["17", TraditionalForm]}, {18, 
       FormBox["18", TraditionalForm]}, {19, 
       FormBox["19", TraditionalForm]}, {20, 
       FormBox["20", TraditionalForm]}, {21, 
       FormBox["21", TraditionalForm]}, {22, 
       FormBox["22", TraditionalForm]}, {23, 
       FormBox["23", TraditionalForm]}, {24, 
       FormBox["24", TraditionalForm]}, {25, 
       FormBox["25", TraditionalForm]}, {26, 
       FormBox["26", TraditionalForm]}, {27, 
       FormBox["27", TraditionalForm]}, {28, 
       FormBox["28", TraditionalForm]}, {29, 
       FormBox["29", TraditionalForm]}, {30, 
       FormBox["30", TraditionalForm]}, {31, 
       FormBox["31", TraditionalForm]}, {32, 
       FormBox["32", TraditionalForm]}, {33, 
       FormBox["33", TraditionalForm]}, {34, 
       FormBox["34", TraditionalForm]}, {35, 
       FormBox["35", TraditionalForm]}, {36, 
       FormBox["36", TraditionalForm]}, {37, 
       FormBox["37", TraditionalForm]}, {38, 
       FormBox["38", TraditionalForm]}, {39, 
       FormBox["39", TraditionalForm]}, {40, 
       FormBox["40", TraditionalForm]}, {41, 
       FormBox["41", TraditionalForm]}, {42, 
       FormBox["42", TraditionalForm]}, {43, 
       FormBox["43", TraditionalForm]}, {44, 
       FormBox["44", TraditionalForm]}, {45, 
       FormBox["45", TraditionalForm]}, {46, 
       FormBox["46", TraditionalForm]}, {47, 
       FormBox["47", TraditionalForm]}, {48, 
       FormBox["48", TraditionalForm]}, {49, 
       FormBox["49", TraditionalForm]}, {50, 
       FormBox["50", TraditionalForm]}, {51, 
       FormBox["51", TraditionalForm]}, {52, 
       FormBox["52", TraditionalForm]}, {53, 
       FormBox["53", TraditionalForm]}, {54, 
       FormBox["54", TraditionalForm]}, {55, 
       FormBox["55", TraditionalForm]}, {56, 
       FormBox["56", TraditionalForm]}, {57, 
       FormBox["57", TraditionalForm]}, {58, 
       FormBox["58", TraditionalForm]}, {59, 
       FormBox["59", TraditionalForm]}, {60, 
       FormBox["60", TraditionalForm]}, {61, 
       FormBox["61", TraditionalForm]}, {62, 
       FormBox["62", TraditionalForm]}, {63, 
       FormBox["63", TraditionalForm]}, {64, 
       FormBox["64", TraditionalForm]}, {65, 
       FormBox["65", TraditionalForm]}, {66, 
       FormBox["66", TraditionalForm]}, {67, 
       FormBox["67", TraditionalForm]}, {68, 
       FormBox["68", TraditionalForm]}, {69, 
       FormBox["69", TraditionalForm]}, {70, 
       FormBox["70", TraditionalForm]}, {71, 
       FormBox["71", TraditionalForm]}, {72, 
       FormBox["72", TraditionalForm]}, {73, 
       FormBox["73", TraditionalForm]}, {74, 
       FormBox["74", TraditionalForm]}, {75, 
       FormBox["75", TraditionalForm]}, {76, 
       FormBox["76", TraditionalForm]}, {77, 
       FormBox["77", TraditionalForm]}, {78, 
       FormBox["78", TraditionalForm]}, {79, 
       FormBox["79", TraditionalForm]}, {80, 
       FormBox["80", TraditionalForm]}, {81, 
       FormBox["81", TraditionalForm]}, {82, 
       FormBox["82", TraditionalForm]}, {83, 
       FormBox["83", TraditionalForm]}, {84, 
       FormBox["84", TraditionalForm]}, {85, 
       FormBox["85", TraditionalForm]}, {86, 
       FormBox["86", TraditionalForm]}, {87, 
       FormBox["87", TraditionalForm]}, {88, 
       FormBox["88", TraditionalForm]}, {89, 
       FormBox["89", TraditionalForm]}, {90, 
       FormBox["90", TraditionalForm]}, {91, 
       FormBox["91", TraditionalForm]}, {92, 
       FormBox["92", TraditionalForm]}, {93, 
       FormBox["93", TraditionalForm]}, {94, 
       FormBox["94", TraditionalForm]}, {95, 
       FormBox["95", TraditionalForm]}, {96, 
       FormBox["96", TraditionalForm]}, {97, 
       FormBox["97", TraditionalForm]}, {98, 
       FormBox["98", TraditionalForm]}, {99, 
       FormBox["99", TraditionalForm]}, {100, 
       FormBox["100", TraditionalForm]}}, {{-100, 
       FormBox["\"\"", TraditionalForm]}, {-99, 
       FormBox["\"\"", TraditionalForm]}, {-98, 
       FormBox["\"\"", TraditionalForm]}, {-97, 
       FormBox["\"\"", TraditionalForm]}, {-96, 
       FormBox["\"\"", TraditionalForm]}, {-95, 
       FormBox["\"\"", TraditionalForm]}, {-94, 
       FormBox["\"\"", TraditionalForm]}, {-93, 
       FormBox["\"\"", TraditionalForm]}, {-92, 
       FormBox["\"\"", TraditionalForm]}, {-91, 
       FormBox["\"\"", TraditionalForm]}, {-90, 
       FormBox["\"\"", TraditionalForm]}, {-89, 
       FormBox["\"\"", TraditionalForm]}, {-88, 
       FormBox["\"\"", TraditionalForm]}, {-87, 
       FormBox["\"\"", TraditionalForm]}, {-86, 
       FormBox["\"\"", TraditionalForm]}, {-85, 
       FormBox["\"\"", TraditionalForm]}, {-84, 
       FormBox["\"\"", TraditionalForm]}, {-83, 
       FormBox["\"\"", TraditionalForm]}, {-82, 
       FormBox["\"\"", TraditionalForm]}, {-81, 
       FormBox["\"\"", TraditionalForm]}, {-80, 
       FormBox["\"\"", TraditionalForm]}, {-79, 
       FormBox["\"\"", TraditionalForm]}, {-78, 
       FormBox["\"\"", TraditionalForm]}, {-77, 
       FormBox["\"\"", TraditionalForm]}, {-76, 
       FormBox["\"\"", TraditionalForm]}, {-75, 
       FormBox["\"\"", TraditionalForm]}, {-74, 
       FormBox["\"\"", TraditionalForm]}, {-73, 
       FormBox["\"\"", TraditionalForm]}, {-72, 
       FormBox["\"\"", TraditionalForm]}, {-71, 
       FormBox["\"\"", TraditionalForm]}, {-70, 
       FormBox["\"\"", TraditionalForm]}, {-69, 
       FormBox["\"\"", TraditionalForm]}, {-68, 
       FormBox["\"\"", TraditionalForm]}, {-67, 
       FormBox["\"\"", TraditionalForm]}, {-66, 
       FormBox["\"\"", TraditionalForm]}, {-65, 
       FormBox["\"\"", TraditionalForm]}, {-64, 
       FormBox["\"\"", TraditionalForm]}, {-63, 
       FormBox["\"\"", TraditionalForm]}, {-62, 
       FormBox["\"\"", TraditionalForm]}, {-61, 
       FormBox["\"\"", TraditionalForm]}, {-60, 
       FormBox["\"\"", TraditionalForm]}, {-59, 
       FormBox["\"\"", TraditionalForm]}, {-58, 
       FormBox["\"\"", TraditionalForm]}, {-57, 
       FormBox["\"\"", TraditionalForm]}, {-56, 
       FormBox["\"\"", TraditionalForm]}, {-55, 
       FormBox["\"\"", TraditionalForm]}, {-54, 
       FormBox["\"\"", TraditionalForm]}, {-53, 
       FormBox["\"\"", TraditionalForm]}, {-52, 
       FormBox["\"\"", TraditionalForm]}, {-51, 
       FormBox["\"\"", TraditionalForm]}, {-50, 
       FormBox["\"\"", TraditionalForm]}, {-49, 
       FormBox["\"\"", TraditionalForm]}, {-48, 
       FormBox["\"\"", TraditionalForm]}, {-47, 
       FormBox["\"\"", TraditionalForm]}, {-46, 
       FormBox["\"\"", TraditionalForm]}, {-45, 
       FormBox["\"\"", TraditionalForm]}, {-44, 
       FormBox["\"\"", TraditionalForm]}, {-43, 
       FormBox["\"\"", TraditionalForm]}, {-42, 
       FormBox["\"\"", TraditionalForm]}, {-41, 
       FormBox["\"\"", TraditionalForm]}, {-40, 
       FormBox["\"\"", TraditionalForm]}, {-39, 
       FormBox["\"\"", TraditionalForm]}, {-38, 
       FormBox["\"\"", TraditionalForm]}, {-37, 
       FormBox["\"\"", TraditionalForm]}, {-36, 
       FormBox["\"\"", TraditionalForm]}, {-35, 
       FormBox["\"\"", TraditionalForm]}, {-34, 
       FormBox["\"\"", TraditionalForm]}, {-33, 
       FormBox["\"\"", TraditionalForm]}, {-32, 
       FormBox["\"\"", TraditionalForm]}, {-31, 
       FormBox["\"\"", TraditionalForm]}, {-30, 
       FormBox["\"\"", TraditionalForm]}, {-29, 
       FormBox["\"\"", TraditionalForm]}, {-28, 
       FormBox["\"\"", TraditionalForm]}, {-27, 
       FormBox["\"\"", TraditionalForm]}, {-26, 
       FormBox["\"\"", TraditionalForm]}, {-25, 
       FormBox["\"\"", TraditionalForm]}, {-24, 
       FormBox["\"\"", TraditionalForm]}, {-23, 
       FormBox["\"\"", TraditionalForm]}, {-22, 
       FormBox["\"\"", TraditionalForm]}, {-21, 
       FormBox["\"\"", TraditionalForm]}, {-20, 
       FormBox["\"\"", TraditionalForm]}, {-19, 
       FormBox["\"\"", TraditionalForm]}, {-18, 
       FormBox["\"\"", TraditionalForm]}, {-17, 
       FormBox["\"\"", TraditionalForm]}, {-16, 
       FormBox["\"\"", TraditionalForm]}, {-15, 
       FormBox["\"\"", TraditionalForm]}, {-14, 
       FormBox["\"\"", TraditionalForm]}, {-13, 
       FormBox["\"\"", TraditionalForm]}, {-12, 
       FormBox["\"\"", TraditionalForm]}, {-11, 
       FormBox["\"\"", TraditionalForm]}, {-10, 
       FormBox["\"\"", TraditionalForm]}, {-9, 
       FormBox["\"\"", TraditionalForm]}, {-8, 
       FormBox["\"\"", TraditionalForm]}, {-7, 
       FormBox["\"\"", TraditionalForm]}, {-6, 
       FormBox["\"\"", TraditionalForm]}, {-5, 
       FormBox["\"\"", TraditionalForm]}, {-4, 
       FormBox["\"\"", TraditionalForm]}, {-3, 
       FormBox["\"\"", TraditionalForm]}, {-2, 
       FormBox["\"\"", TraditionalForm]}, {-1, 
       FormBox["\"\"", TraditionalForm]}, {0, 
       FormBox["\"\"", TraditionalForm]}, {1, 
       FormBox["\"\"", TraditionalForm]}, {2, 
       FormBox["\"\"", TraditionalForm]}, {3, 
       FormBox["\"\"", TraditionalForm]}, {4, 
       FormBox["\"\"", TraditionalForm]}, {5, 
       FormBox["\"\"", TraditionalForm]}, {6, 
       FormBox["\"\"", TraditionalForm]}, {7, 
       FormBox["\"\"", TraditionalForm]}, {8, 
       FormBox["\"\"", TraditionalForm]}, {9, 
       FormBox["\"\"", TraditionalForm]}, {10, 
       FormBox["\"\"", TraditionalForm]}, {11, 
       FormBox["\"\"", TraditionalForm]}, {12, 
       FormBox["\"\"", TraditionalForm]}, {13, 
       FormBox["\"\"", TraditionalForm]}, {14, 
       FormBox["\"\"", TraditionalForm]}, {15, 
       FormBox["\"\"", TraditionalForm]}, {16, 
       FormBox["\"\"", TraditionalForm]}, {17, 
       FormBox["\"\"", TraditionalForm]}, {18, 
       FormBox["\"\"", TraditionalForm]}, {19, 
       FormBox["\"\"", TraditionalForm]}, {20, 
       FormBox["\"\"", TraditionalForm]}, {21, 
       FormBox["\"\"", TraditionalForm]}, {22, 
       FormBox["\"\"", TraditionalForm]}, {23, 
       FormBox["\"\"", TraditionalForm]}, {24, 
       FormBox["\"\"", TraditionalForm]}, {25, 
       FormBox["\"\"", TraditionalForm]}, {26, 
       FormBox["\"\"", TraditionalForm]}, {27, 
       FormBox["\"\"", TraditionalForm]}, {28, 
       FormBox["\"\"", TraditionalForm]}, {29, 
       FormBox["\"\"", TraditionalForm]}, {30, 
       FormBox["\"\"", TraditionalForm]}, {31, 
       FormBox["\"\"", TraditionalForm]}, {32, 
       FormBox["\"\"", TraditionalForm]}, {33, 
       FormBox["\"\"", TraditionalForm]}, {34, 
       FormBox["\"\"", TraditionalForm]}, {35, 
       FormBox["\"\"", TraditionalForm]}, {36, 
       FormBox["\"\"", TraditionalForm]}, {37, 
       FormBox["\"\"", TraditionalForm]}, {38, 
       FormBox["\"\"", TraditionalForm]}, {39, 
       FormBox["\"\"", TraditionalForm]}, {40, 
       FormBox["\"\"", TraditionalForm]}, {41, 
       FormBox["\"\"", TraditionalForm]}, {42, 
       FormBox["\"\"", TraditionalForm]}, {43, 
       FormBox["\"\"", TraditionalForm]}, {44, 
       FormBox["\"\"", TraditionalForm]}, {45, 
       FormBox["\"\"", TraditionalForm]}, {46, 
       FormBox["\"\"", TraditionalForm]}, {47, 
       FormBox["\"\"", TraditionalForm]}, {48, 
       FormBox["\"\"", TraditionalForm]}, {49, 
       FormBox["\"\"", TraditionalForm]}, {50, 
       FormBox["\"\"", TraditionalForm]}, {51, 
       FormBox["\"\"", TraditionalForm]}, {52, 
       FormBox["\"\"", TraditionalForm]}, {53, 
       FormBox["\"\"", TraditionalForm]}, {54, 
       FormBox["\"\"", TraditionalForm]}, {55, 
       FormBox["\"\"", TraditionalForm]}, {56, 
       FormBox["\"\"", TraditionalForm]}, {57, 
       FormBox["\"\"", TraditionalForm]}, {58, 
       FormBox["\"\"", TraditionalForm]}, {59, 
       FormBox["\"\"", TraditionalForm]}, {60, 
       FormBox["\"\"", TraditionalForm]}, {61, 
       FormBox["\"\"", TraditionalForm]}, {62, 
       FormBox["\"\"", TraditionalForm]}, {63, 
       FormBox["\"\"", TraditionalForm]}, {64, 
       FormBox["\"\"", TraditionalForm]}, {65, 
       FormBox["\"\"", TraditionalForm]}, {66, 
       FormBox["\"\"", TraditionalForm]}, {67, 
       FormBox["\"\"", TraditionalForm]}, {68, 
       FormBox["\"\"", TraditionalForm]}, {69, 
       FormBox["\"\"", TraditionalForm]}, {70, 
       FormBox["\"\"", TraditionalForm]}, {71, 
       FormBox["\"\"", TraditionalForm]}, {72, 
       FormBox["\"\"", TraditionalForm]}, {73, 
       FormBox["\"\"", TraditionalForm]}, {74, 
       FormBox["\"\"", TraditionalForm]}, {75, 
       FormBox["\"\"", TraditionalForm]}, {76, 
       FormBox["\"\"", TraditionalForm]}, {77, 
       FormBox["\"\"", TraditionalForm]}, {78, 
       FormBox["\"\"", TraditionalForm]}, {79, 
       FormBox["\"\"", TraditionalForm]}, {80, 
       FormBox["\"\"", TraditionalForm]}, {81, 
       FormBox["\"\"", TraditionalForm]}, {82, 
       FormBox["\"\"", TraditionalForm]}, {83, 
       FormBox["\"\"", TraditionalForm]}, {84, 
       FormBox["\"\"", TraditionalForm]}, {85, 
       FormBox["\"\"", TraditionalForm]}, {86, 
       FormBox["\"\"", TraditionalForm]}, {87, 
       FormBox["\"\"", TraditionalForm]}, {88, 
       FormBox["\"\"", TraditionalForm]}, {89, 
       FormBox["\"\"", TraditionalForm]}, {90, 
       FormBox["\"\"", TraditionalForm]}, {91, 
       FormBox["\"\"", TraditionalForm]}, {92, 
       FormBox["\"\"", TraditionalForm]}, {93, 
       FormBox["\"\"", TraditionalForm]}, {94, 
       FormBox["\"\"", TraditionalForm]}, {95, 
       FormBox["\"\"", TraditionalForm]}, {96, 
       FormBox["\"\"", TraditionalForm]}, {97, 
       FormBox["\"\"", TraditionalForm]}, {98, 
       FormBox["\"\"", TraditionalForm]}, {99, 
       FormBox["\"\"", TraditionalForm]}, {100, 
       FormBox["\"\"", TraditionalForm]}}}, {{{2, 
       FormBox["2", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["3", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["4", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["5", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["6", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["7", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["8", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["9", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["10", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["11", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["12", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["13", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["14", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["15", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["16", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["17", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["18", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["19", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}, {{2, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}}},
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImageSize->400,
  Method->{"DefaultBoundaryStyle" -> Automatic, "ScalingFunctions" -> None},
  PlotLabel->FormBox["\"\"", TraditionalForm],
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.05], 
     Scaled[0.05]}},
  Prolog->{{{
      GrayLevel[0.9], 
      RectangleBox[{10.561453473142157`, -10000}, {
       12.222494362608364`, 10000}]}, {
      GrayLevel[0.8], 
      RectangleBox[{12.222494362608364`, -10000}, {
       14.334981351900138`, 10000}]}, {
      GrayLevel[0.7], 
      RectangleBox[{14.334981351900138`, -10000}, {17, 10000}]}}},
  Ticks->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.636847125276335*^9, 3.636847466767321*^9, 
  3.646456148217771*^9, 3.646475893028215*^9, 3.646477154236602*^9, 
  3.646477806556148*^9}]
}, Open  ]],

Cell["\<\
Draw the plot for the atmospheric mass squared difference (on a linear scale)\
\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"Plot", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{"\[CapitalDelta]m2atm", "[", "t", "]"}], "*", 
      RowBox[{"10", "^", "3"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", "tmin", ",", "tmax"}], "}"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"ImageSize", "\[Rule]", "400"}], ",", 
     RowBox[{"FrameLabel", "->", 
      RowBox[{"{", 
       RowBox[{
       "\"\<\!\(\*SubscriptBox[\(log\), \(10\)]\)(\[Mu]/GeV)\>\"", ",", 
        "\"\<\!\(\*SubsuperscriptBox[\(\[CapitalDelta]m\), \(atm\), \(2\)]\) \
[\!\(\*SuperscriptBox[\(10\), \
\(-3\)]\)\[ThinSpace]\!\(\*SuperscriptBox[\(eV\), \(2\)]\)]\>\""}], "}"}]}], 
     ",", 
     RowBox[{"FrameTicks", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"RGELogTicksLabeled", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"i", "*", "0.2"}], ",", 
            RowBox[{"PaddedForm", "[", 
             RowBox[{
              RowBox[{"i", "*", "0.2"}], ",", 
              RowBox[{"{", 
               RowBox[{"2", ",", "1"}], "}"}]}], "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", 
            RowBox[{"-", "50"}], ",", "50"}], "}"}]}], "]"}], ",", 
        RowBox[{"RGELogTicks", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", 
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{"i", "*", "0.2"}], ",", "\"\<\>\""}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", "0", ",", "50"}], "}"}]}], "]"}]}], "}"}]}], ",",
      "\[IndentingNewLine]", 
     RowBox[{"PlotRange", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"All", ",", "All"}], "}"}]}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"{", "Blue", "}"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"Prolog", "\[Rule]", 
      RowBox[{"{", "ShadowEFT", "}"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"DisplayFunction", "\[Rule]", "Identity"}]}], 
    "\[IndentingNewLine]", "]"}], ",", 
   RowBox[{"DisplayFunction", "\[Rule]", "$DisplayFunction"}]}], 
  "]"}]], "Input",
 CellChangeTimes->{3.435655660381*^9, 3.435655690904253*^9}],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {RGBColor[0, 0, 1], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwV03k01Psfx/Ex3zEzwnd00VXKvvtiiiiK15UaRqUstwi5IdwGpbQIpXJb
LJU90SJCSpS2W7KXpRHSokWoKKFJKdHV7/v743M+53HO53w+z/M+56O5Idw1
kMlgMPLo9f/92ZZSyzSvX3bXyz3zt91l4JDgtmeW6Jdd9ViE5rUaBqzmNEbn
xP6yqw/0osbqGcho7K3Nz/tl586eMbS1hQF3dWWXa4O/7KT8n88QPWMguL/S
fNNiBn7OYFkJRhl4m6qzPFHMQIdLYkqxrhRejHyObh2Wgs7XkxW6CVLodyuw
aPdlQW9T+ubEv5iYH6mU+imIBemvJdNPBzBxIGP/Z/ktLES0flxeHsSEZtdf
pcL9LHxe4JzSGcqEt6+aQX0hC+mnBetmRDHxKDhT9YaEhQW1dhP/pDBRHXOY
mbtPGt+tuZw5dUzwTo/73U6QhrapZKb2PSZ8q4OqutKkUZDypMmgiYlJ5rLo
GYXSCJn3iWH+kAnLQ8xvR1ukUeV8uNX+BROXUnZ/2KfEhhs//ofLFyayC0Pb
gs+z4ZVkpaWpTWBiu1eA52U2ahpnycnpEfBcJhh3uskGJ3vZiW8GBFTeaWgY
N7MhZ1yb22xKIF2rM3x4mA1nO4/VYdYEjuVa8yLmc3D5+66t+asJSET65zbY
ciC6UZR92J2AyyIlKzcBB9XBDhvD1hAgXwz7WnhyoHWX89HSh0CCytnSb9Ec
3Ko8f7EumMA/qZyVu+s58HO/qf5gD4H+DV97N4k5uJI9kF6yj8DSeb2R3k84
sFl1ovxIPAFW+7+5i99z4JXd2CVIIOB/pnCu6WcOdH4E9esmE6gNT2tQm+Cg
VCTOIo4T0LKL8+QRXCQ9GRRGpRKII8OGf8lyERa5JMU4nUDPK684iRIXwsml
dS8zCOCSQLl3DhcVqyWvkrIITDlrLq4148Jb1aLt00kC0Yc7Ew+4cdEjZ1t0
JY/Ay7U1GpHeXEyfcNDzzydgY1BaERjIBcvoTZLieXre9w6+XLaDC676q5nb
ignsZNlQMie56LBI1rEoI7B1z9mWpD4uShaqE2WVBJLL+9diiIt4B9e8/VUE
Lrwx7h8d48K0hKG/poZAn+A6Y62MDHZtauj5r56Aq0LLfC2+DFR0i5SFDwjw
z349fT1GBjfE1ftePScwUivY9lplGp7VyfhKvhMIZo/Mnusii6ndJtsb5rLA
jnZxDUmRR6Cpx5XUHBb82t6O8AgFFI7V2uaS0niYv+btXKvpWLo9sPRBtDRS
JZ3Op/78DWuF89K+9krjXapvt1OyIrKf/MzqXMFGQulon125EmaGqJxpLmPD
IWFRZnSrMsaCF7/QUOAgoSpl+zjxO2aa80IKIjn0v/py/66OCnxcthnHd3Dg
EubnUO8zE3eDtLrU53ORYL5/s8LxWfi5JeDx/VQuZm8uIwsvqyKQ4zRbZZAL
a9bg0+j+2Xh6J3HOgaUyuMQzVczlqqG8JXPXnSwZlEbE7vppr44xx6m2JIkM
ep1rBe02GnB9FuPo7jANytnNCU0tGgjy26Jnd2waFJQ8A3uDNFEWPmDP65mG
a01DO0790MRfJSxXEwNZBPit64s9pAXvrck2qSJZPNda1RerrY3jxYp12hWy
mNAQqupUaGPCzI0T8FUWvcLmD+9X6WBnXXa9uZUc2j10FdoHdfCSKVvZHSkH
UU58uG+ULsyWR+oVXpKDmBQFyM7Qw2Hp34uNB+TQdWdNik+hHv5xi/KQmiWP
EWc1lUfz9TF5fOHmWU7yaGZvMg9u08fVNR/Yyrvkseq/Owfz/Aygxjlb35Un
j1z19C0/fxpgbqL8h0NieaTHRKRNJhtiR53ZO+dReWTcMeKpqRlhpgoKhIok
fPVD1P1uGeGl/4vPJ8xJtIkvBguExqizV2WGriQhc2a7YOWAMbJ21iQ0B5F4
Mm+15PoHY7jKnl+vF0wiv8H4hPqQMc4JB77F0bb92DMokRij5pkofEEIiQhL
5+TUCWNMFf46mPc3iecP1B8/k6cQW67ltimUxIUfjRv8LSiYju/jP4wgsSPx
nOwDSwreHd0btbeScFCPrbBYSOHJeim/HbS7HczZbFsKSufjstW2kfjtWG5x
kYDC3qJHZn9HkojSi5AMe1EIfr/96dgOEs5uqrE74yhY/3tzvDuaBHuy6/aN
/RQys5y99WJI1OZl/fgWT0HLmDwSStt6VDky8ggFjVm/OUzSNjiuIIpIpeC0
9fxcpT0k3i54WFyeTmFcS9DvSftMT9KAJJNCTtP3gNO0f+fL+ofnUPBYtKva
cC+JjqdNZ0pPUWjXtM8No52851D38BkKRZayDldpsx9Ke4kKKGwwTP5kE0f3
bK/PLCmk4K+Nr3tox6rtfzxYTOFcfF9tHW3re38oGl2kwBfHnrbZR2IslLE6
pJRCw9TeC/P2kyhTrkouKqNg+cYj3fAACVFlzIOBKxQukuPrNeLpfrlJwcYb
FByTktPkD9L9FbfiC25ROC+zc4o4RMLbe2fd29sUdDVmOk7QflQyZutfTeFu
uCiq/wjdJ5TM92ui7/9S5VWZTPeNlm493UJhUVZrzZWjdF92aHm3mMLggRBu
0TES5R8GKZ8OCheSTS1SUuj3Dr3T9npBoa/PRzowg0TMvefT3YcpdClkHGad
IrHAvmFYMN0EpzJPfDG6ROIc83nxDUUT3L/6OJNfSoKsGwnUn2GCFWlqypaX
SbxZotLNVjVBvO7tbJTT8126qbVBxwTWZuvb3a6ReOc4/fKSBSa4OrSqP6KS
RIqLz2a79Sa4vKQrNlVMYsh7bNTykglaOtw/eg+TEF41mttia4pnF7ZdZujz
UD5ofX+f2BTjlQ8Iex8eXq9oH/UMNsPIdOeQ3ek83BwSrXkyboaAtD1b1t3n
IYrlunfdpBlsHPSFFo08LJ5tVdzznxnM0zNG5Jp46C3YGxMpxUfYybzsymYe
Nj4WPtzD5sPp0YrY2a08bLZ4HZGmwAchfP13WycP8aOcfyt1+FipUHJUoY8+
Hxu1kqHHxyVTQtxH23HacJ+9Ph/OJ9s6K97wME2rQ7bRkI/QP88sWfuOh2Or
c3w6TPmontefk/Oeh+wyvtTAAj5G9fMsVD/xsHvRuXRDaz7W99orDdL2blQ2
EtnwkdNWUnlTwoN6z4Tr58V8vD56IdRjlId8XkP+pD0ft0IM/0gco/tOWi20
deCDnd8V6vmN7tO/IN67lA/XuAl/ve88GNod/SbtyMcIa8PFqnG6r0UqwdGJ
D+m8lKmEHzwM/blNPUHIx/cWzqy1EzyI+/qvip358N9xfUp7kofSME9HhRV8
LE9PvfiJ9v8A/NxCww==
     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{False, False},
  AxesLabel->{None, None},
  AxesOrigin->{2., 2.},
  DisplayFunction->Identity,
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{
     FormBox[
     "\"\\!\\(\\*SubsuperscriptBox[\\(\[CapitalDelta]m\\), \\(atm\\), \
\\(2\\)]\\) [\\!\\(\\*SuperscriptBox[\\(10\\), \\(-3\\)]\\)\[ThinSpace]\\!\\(\
\\*SuperscriptBox[\\(eV\\), \\(2\\)]\\)]\"", TraditionalForm], None}, {
     FormBox[
     "\"\\!\\(\\*SubscriptBox[\\(log\\), \\(10\\)]\\)(\[Mu]/GeV)\"", 
      TraditionalForm], None}},
  FrameTicks->{{{{-10., 
       FormBox[
        TagBox[
         InterpretationBox["\"-10.0\"", -10., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-9.8, 
       FormBox[
        TagBox[
         InterpretationBox["\"-9.8\"", -9.8, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-9.600000000000001, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-9.6\"", -9.600000000000001, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-9.4, 
       FormBox[
        TagBox[
         InterpretationBox["\"-9.4\"", -9.4, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-9.200000000000001, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-9.2\"", -9.200000000000001, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-9., 
       FormBox[
        TagBox[
         InterpretationBox["\"-9.0\"", -9., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-8.8, 
       FormBox[
        TagBox[
         InterpretationBox["\"-8.8\"", -8.8, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-8.6, 
       FormBox[
        TagBox[
         InterpretationBox["\"-8.6\"", -8.6, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-8.4, 
       FormBox[
        TagBox[
         InterpretationBox["\"-8.4\"", -8.4, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-8.200000000000001, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-8.2\"", -8.200000000000001, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-8., 
       FormBox[
        TagBox[
         InterpretationBox["\"-8.0\"", -8., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-7.800000000000001, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-7.8\"", -7.800000000000001, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-7.6000000000000005`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-7.6\"", -7.6000000000000005`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-7.4, 
       FormBox[
        TagBox[
         InterpretationBox["\"-7.4\"", -7.4, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-7.2, 
       FormBox[
        TagBox[
         InterpretationBox["\"-7.2\"", -7.2, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-7., 
       FormBox[
        TagBox[
         InterpretationBox["\"-7.0\"", -7., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-6.800000000000001, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-6.8\"", -6.800000000000001, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-6.6000000000000005`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-6.6\"", -6.6000000000000005`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-6.4, 
       FormBox[
        TagBox[
         InterpretationBox["\"-6.4\"", -6.4, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-6.2, 
       FormBox[
        TagBox[
         InterpretationBox["\"-6.2\"", -6.2, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-6., 
       FormBox[
        TagBox[
         InterpretationBox["\"-6.0\"", -6., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-5.800000000000001, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-5.8\"", -5.800000000000001, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-5.6000000000000005`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-5.6\"", -5.6000000000000005`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-5.4, 
       FormBox[
        TagBox[
         InterpretationBox["\"-5.4\"", -5.4, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-5.2, 
       FormBox[
        TagBox[
         InterpretationBox["\"-5.2\"", -5.2, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-5., 
       FormBox[
        TagBox[
         InterpretationBox["\"-5.0\"", -5., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-4.800000000000001, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-4.8\"", -4.800000000000001, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-4.6000000000000005`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-4.6\"", -4.6000000000000005`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-4.4, 
       FormBox[
        TagBox[
         InterpretationBox["\"-4.4\"", -4.4, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-4.2, 
       FormBox[
        TagBox[
         InterpretationBox["\"-4.2\"", -4.2, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-4., 
       FormBox[
        TagBox[
         InterpretationBox["\"-4.0\"", -4., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-3.8000000000000003`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-3.8\"", -3.8000000000000003`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-3.6, 
       FormBox[
        TagBox[
         InterpretationBox["\"-3.6\"", -3.6, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-3.4000000000000004`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-3.4\"", -3.4000000000000004`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-3.2, 
       FormBox[
        TagBox[
         InterpretationBox["\"-3.2\"", -3.2, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-3., 
       FormBox[
        TagBox[
         InterpretationBox["\"-3.0\"", -3., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-2.8000000000000003`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-2.8\"", -2.8000000000000003`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-2.6, 
       FormBox[
        TagBox[
         InterpretationBox["\"-2.6\"", -2.6, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-2.4000000000000004`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-2.4\"", -2.4000000000000004`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-2.2, 
       FormBox[
        TagBox[
         InterpretationBox["\"-2.2\"", -2.2, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-2., 
       FormBox[
        TagBox[
         InterpretationBox["\"-2.0\"", -2., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-1.8, 
       FormBox[
        TagBox[
         InterpretationBox["\"-1.8\"", -1.8, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-1.6, 
       FormBox[
        TagBox[
         InterpretationBox["\"-1.6\"", -1.6, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-1.4000000000000001`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-1.4\"", -1.4000000000000001`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-1.2000000000000002`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-1.2\"", -1.2000000000000002`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-1., 
       FormBox[
        TagBox[
         InterpretationBox["\"-1.0\"", -1., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-0.8, 
       FormBox[
        TagBox[
         InterpretationBox["\"-0.8\"", -0.8, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-0.6000000000000001, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\"-0.6\"", -0.6000000000000001, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-0.4, 
       FormBox[
        TagBox[
         InterpretationBox["\"-0.4\"", -0.4, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {-0.2, 
       FormBox[
        TagBox[
         InterpretationBox["\"-0.2\"", -0.2, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {0., 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.0\"", 0., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {0.2, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.2\"", 0.2, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {0.4, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.4\"", 0.4, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {0.6000000000000001, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\" 0.6\"", 0.6000000000000001, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {0.8, 
       FormBox[
        TagBox[
         InterpretationBox["\" 0.8\"", 0.8, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {1., 
       FormBox[
        TagBox[
         InterpretationBox["\" 1.0\"", 1., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {1.2000000000000002`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\" 1.2\"", 1.2000000000000002`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {1.4000000000000001`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\" 1.4\"", 1.4000000000000001`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {1.6, 
       FormBox[
        TagBox[
         InterpretationBox["\" 1.6\"", 1.6, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {1.8, 
       FormBox[
        TagBox[
         InterpretationBox["\" 1.8\"", 1.8, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {2., 
       FormBox[
        TagBox[
         InterpretationBox["\" 2.0\"", 2., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {2.2, 
       FormBox[
        TagBox[
         InterpretationBox["\" 2.2\"", 2.2, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {2.4000000000000004`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\" 2.4\"", 2.4000000000000004`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {2.6, 
       FormBox[
        TagBox[
         InterpretationBox["\" 2.6\"", 2.6, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {2.8000000000000003`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\" 2.8\"", 2.8000000000000003`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {3., 
       FormBox[
        TagBox[
         InterpretationBox["\" 3.0\"", 3., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {3.2, 
       FormBox[
        TagBox[
         InterpretationBox["\" 3.2\"", 3.2, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {3.4000000000000004`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\" 3.4\"", 3.4000000000000004`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {3.6, 
       FormBox[
        TagBox[
         InterpretationBox["\" 3.6\"", 3.6, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {3.8000000000000003`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\" 3.8\"", 3.8000000000000003`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {4., 
       FormBox[
        TagBox[
         InterpretationBox["\" 4.0\"", 4., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {4.2, 
       FormBox[
        TagBox[
         InterpretationBox["\" 4.2\"", 4.2, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {4.4, 
       FormBox[
        TagBox[
         InterpretationBox["\" 4.4\"", 4.4, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {4.6000000000000005`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\" 4.6\"", 4.6000000000000005`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {4.800000000000001, 
       FormBox[
        TagBox[
         InterpretationBox["\" 4.8\"", 4.800000000000001, AutoDelete -> True],
          PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {5., 
       FormBox[
        TagBox[
         InterpretationBox["\" 5.0\"", 5., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {5.2, 
       FormBox[
        TagBox[
         InterpretationBox["\" 5.2\"", 5.2, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {5.4, 
       FormBox[
        TagBox[
         InterpretationBox["\" 5.4\"", 5.4, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {5.6000000000000005`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\" 5.6\"", 5.6000000000000005`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {5.800000000000001, 
       FormBox[
        TagBox[
         InterpretationBox["\" 5.8\"", 5.800000000000001, AutoDelete -> True],
          PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {6., 
       FormBox[
        TagBox[
         InterpretationBox["\" 6.0\"", 6., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {6.2, 
       FormBox[
        TagBox[
         InterpretationBox["\" 6.2\"", 6.2, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {6.4, 
       FormBox[
        TagBox[
         InterpretationBox["\" 6.4\"", 6.4, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {6.6000000000000005`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\" 6.6\"", 6.6000000000000005`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {6.800000000000001, 
       FormBox[
        TagBox[
         InterpretationBox["\" 6.8\"", 6.800000000000001, AutoDelete -> True],
          PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {7., 
       FormBox[
        TagBox[
         InterpretationBox["\" 7.0\"", 7., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {7.2, 
       FormBox[
        TagBox[
         InterpretationBox["\" 7.2\"", 7.2, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {7.4, 
       FormBox[
        TagBox[
         InterpretationBox["\" 7.4\"", 7.4, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {7.6000000000000005`, 
       FormBox[
        TagBox[
         InterpretationBox[
         "\" 7.6\"", 7.6000000000000005`, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {7.800000000000001, 
       FormBox[
        TagBox[
         InterpretationBox["\" 7.8\"", 7.800000000000001, AutoDelete -> True],
          PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {8., 
       FormBox[
        TagBox[
         InterpretationBox["\" 8.0\"", 8., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {8.200000000000001, 
       FormBox[
        TagBox[
         InterpretationBox["\" 8.2\"", 8.200000000000001, AutoDelete -> True],
          PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {8.4, 
       FormBox[
        TagBox[
         InterpretationBox["\" 8.4\"", 8.4, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {8.6, 
       FormBox[
        TagBox[
         InterpretationBox["\" 8.6\"", 8.6, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {8.8, 
       FormBox[
        TagBox[
         InterpretationBox["\" 8.8\"", 8.8, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {9., 
       FormBox[
        TagBox[
         InterpretationBox["\" 9.0\"", 9., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {9.200000000000001, 
       FormBox[
        TagBox[
         InterpretationBox["\" 9.2\"", 9.200000000000001, AutoDelete -> True],
          PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {9.4, 
       FormBox[
        TagBox[
         InterpretationBox["\" 9.4\"", 9.4, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {9.600000000000001, 
       FormBox[
        TagBox[
         InterpretationBox["\" 9.6\"", 9.600000000000001, AutoDelete -> True],
          PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {9.8, 
       FormBox[
        TagBox[
         InterpretationBox["\" 9.8\"", 9.8, AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}, {10., 
       FormBox[
        TagBox[
         InterpretationBox["\" 10.0\"", 10., AutoDelete -> True], 
         PaddedForm[#, {2, 1}]& ], TraditionalForm]}}, {{0., 
       FormBox["\"\"", TraditionalForm]}, {0.2, 
       FormBox["\"\"", TraditionalForm]}, {0.4, 
       FormBox["\"\"", TraditionalForm]}, {0.6000000000000001, 
       FormBox["\"\"", TraditionalForm]}, {0.8, 
       FormBox["\"\"", TraditionalForm]}, {1., 
       FormBox["\"\"", TraditionalForm]}, {1.2000000000000002`, 
       FormBox["\"\"", TraditionalForm]}, {1.4000000000000001`, 
       FormBox["\"\"", TraditionalForm]}, {1.6, 
       FormBox["\"\"", TraditionalForm]}, {1.8, 
       FormBox["\"\"", TraditionalForm]}, {2., 
       FormBox["\"\"", TraditionalForm]}, {2.2, 
       FormBox["\"\"", TraditionalForm]}, {2.4000000000000004`, 
       FormBox["\"\"", TraditionalForm]}, {2.6, 
       FormBox["\"\"", TraditionalForm]}, {2.8000000000000003`, 
       FormBox["\"\"", TraditionalForm]}, {3., 
       FormBox["\"\"", TraditionalForm]}, {3.2, 
       FormBox["\"\"", TraditionalForm]}, {3.4000000000000004`, 
       FormBox["\"\"", TraditionalForm]}, {3.6, 
       FormBox["\"\"", TraditionalForm]}, {3.8000000000000003`, 
       FormBox["\"\"", TraditionalForm]}, {4., 
       FormBox["\"\"", TraditionalForm]}, {4.2, 
       FormBox["\"\"", TraditionalForm]}, {4.4, 
       FormBox["\"\"", TraditionalForm]}, {4.6000000000000005`, 
       FormBox["\"\"", TraditionalForm]}, {4.800000000000001, 
       FormBox["\"\"", TraditionalForm]}, {5., 
       FormBox["\"\"", TraditionalForm]}, {5.2, 
       FormBox["\"\"", TraditionalForm]}, {5.4, 
       FormBox["\"\"", TraditionalForm]}, {5.6000000000000005`, 
       FormBox["\"\"", TraditionalForm]}, {5.800000000000001, 
       FormBox["\"\"", TraditionalForm]}, {6., 
       FormBox["\"\"", TraditionalForm]}, {6.2, 
       FormBox["\"\"", TraditionalForm]}, {6.4, 
       FormBox["\"\"", TraditionalForm]}, {6.6000000000000005`, 
       FormBox["\"\"", TraditionalForm]}, {6.800000000000001, 
       FormBox["\"\"", TraditionalForm]}, {7., 
       FormBox["\"\"", TraditionalForm]}, {7.2, 
       FormBox["\"\"", TraditionalForm]}, {7.4, 
       FormBox["\"\"", TraditionalForm]}, {7.6000000000000005`, 
       FormBox["\"\"", TraditionalForm]}, {7.800000000000001, 
       FormBox["\"\"", TraditionalForm]}, {8., 
       FormBox["\"\"", TraditionalForm]}, {8.200000000000001, 
       FormBox["\"\"", TraditionalForm]}, {8.4, 
       FormBox["\"\"", TraditionalForm]}, {8.6, 
       FormBox["\"\"", TraditionalForm]}, {8.8, 
       FormBox["\"\"", TraditionalForm]}, {9., 
       FormBox["\"\"", TraditionalForm]}, {9.200000000000001, 
       FormBox["\"\"", TraditionalForm]}, {9.4, 
       FormBox["\"\"", TraditionalForm]}, {9.600000000000001, 
       FormBox["\"\"", TraditionalForm]}, {9.8, 
       FormBox["\"\"", TraditionalForm]}, {10., 
       FormBox["\"\"", TraditionalForm]}}}, {{{2, 
       FormBox["2", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["3", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["4", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["5", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["6", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["7", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["8", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["9", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["10", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["11", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["12", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["13", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["14", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["15", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["16", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["17", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["18", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["19", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}, {{2, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}}},
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImageSize->400,
  Method->{"DefaultBoundaryStyle" -> Automatic, "ScalingFunctions" -> None},
  PlotLabel->FormBox["\"\"", TraditionalForm],
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.05], 
     Scaled[0.05]}},
  Prolog->{{{
      GrayLevel[0.9], 
      RectangleBox[{10.561453473142157`, -10000}, {
       12.222494362608364`, 10000}]}, {
      GrayLevel[0.8], 
      RectangleBox[{12.222494362608364`, -10000}, {
       14.334981351900138`, 10000}]}, {
      GrayLevel[0.7], 
      RectangleBox[{14.334981351900138`, -10000}, {17, 10000}]}}},
  Ticks->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.63684712581448*^9, 3.636847467341304*^9, 
  3.6464561495243464`*^9, 3.646475893902885*^9, 3.6464771551492777`*^9, 
  3.646477807327272*^9}]
}, Open  ]],

Cell["RG change of the mass squared differences:", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<The RG change of \!\(\*SubscriptBox[SuperscriptBox[\(\[CapitalDelta]m\
\), \(2\)], \(sol\)]\) is:     \!\(\*SubscriptBox[SuperscriptBox[\(\
\[CapitalDelta]m\), \(2\)], \(sol\)]\)(\!\(\*SubscriptBox[\(M\), \(GUT\)]\))-\
\!\(\*SubscriptBox[SuperscriptBox[\(\[CapitalDelta]m\), \(2\)], \(sol\)]\)(\!\
\(\*SubscriptBox[\(M\), \(Z\)]\)) = \>\"", ",", 
    RowBox[{
     RowBox[{"\[CapitalDelta]m2sol", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "GUT"}], "]"}], "]"}], "-", 
     RowBox[{"\[CapitalDelta]m2sol", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "MZ"}], "]"}], "]"}]}], ",", 
    "\"\< \!\(\*SuperscriptBox[\(eV\), \(2\)]\)\>\""}], "]"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{
   "\"\<The RG change of \!\(\*SubscriptBox[SuperscriptBox[\(\[CapitalDelta]m\
\), \(2\)], \(atm\)]\) is:     \!\(\*SubscriptBox[SuperscriptBox[\(\
\[CapitalDelta]m\), \(2\)], \(atm\)]\)(\!\(\*SubscriptBox[\(M\), \(GUT\)]\))-\
\!\(\*SubscriptBox[SuperscriptBox[\(\[CapitalDelta]m\), \(2\)], \(atm\)]\)(\!\
\(\*SubscriptBox[\(M\), \(Z\)]\)) = \>\"", ",", 
    RowBox[{
     RowBox[{"\[CapitalDelta]m2atm", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "GUT"}], "]"}], "]"}], "-", 
     RowBox[{"\[CapitalDelta]m2atm", "[", 
      RowBox[{"Log", "[", 
       RowBox[{"10", ",", "MZ"}], "]"}], "]"}]}], ",", 
    "\"\< \!\(\*SuperscriptBox[\(eV\), \(2\)]\)\>\""}], "]"}], 
  ";"}]}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The RG change of \\!\\(\\*SubscriptBox[SuperscriptBox[\\(\
\[CapitalDelta]m\\), \\(2\\)], \\(sol\\)]\\) is:     \
\\!\\(\\*SubscriptBox[SuperscriptBox[\\(\[CapitalDelta]m\\), \\(2\\)], \\(sol\
\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \
\\(GUT\\)]\\))-\\!\\(\\*SubscriptBox[SuperscriptBox[\\(\[CapitalDelta]m\\), \
\\(2\\)], \\(sol\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\)) = \"\>", 
   "\[InvisibleSpace]", "0.00008281272704692717`", 
   "\[InvisibleSpace]", "\<\" \\!\\(\\*SuperscriptBox[\\(eV\\), \\(2\\)]\\)\"\
\>"}],
  SequenceForm[
  "The RG change of \!\(\*SubscriptBox[SuperscriptBox[\(\[CapitalDelta]m\), \
\(2\)], \(sol\)]\) is:     \
\!\(\*SubscriptBox[SuperscriptBox[\(\[CapitalDelta]m\), \(2\)], \(sol\)]\)(\!\
\(\*SubscriptBox[\(M\), \(GUT\)]\))-\!\(\*SubscriptBox[SuperscriptBox[\(\
\[CapitalDelta]m\), \(2\)], \(sol\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\)) = \
", 0.00008281272704692717, " \!\(\*SuperscriptBox[\(eV\), \(2\)]\)"],
  Editable->False]], "Print",
 CellChangeTimes->{3.63684712587848*^9, 3.6368474673944283`*^9, 
  3.646456149730103*^9, 3.6464758940037413`*^9, 3.646477155341527*^9, 
  3.646477807392768*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"The RG change of \\!\\(\\*SubscriptBox[SuperscriptBox[\\(\
\[CapitalDelta]m\\), \\(2\\)], \\(atm\\)]\\) is:     \
\\!\\(\\*SubscriptBox[SuperscriptBox[\\(\[CapitalDelta]m\\), \\(2\\)], \\(atm\
\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \
\\(GUT\\)]\\))-\\!\\(\\*SubscriptBox[SuperscriptBox[\\(\[CapitalDelta]m\\), \
\\(2\\)], \\(atm\\)]\\)(\\!\\(\\*SubscriptBox[\\(M\\), \\(Z\\)]\\)) = \"\>", 
   "\[InvisibleSpace]", "0.0019050199501955715`", 
   "\[InvisibleSpace]", "\<\" \\!\\(\\*SuperscriptBox[\\(eV\\), \\(2\\)]\\)\"\
\>"}],
  SequenceForm[
  "The RG change of \!\(\*SubscriptBox[SuperscriptBox[\(\[CapitalDelta]m\), \
\(2\)], \(atm\)]\) is:     \
\!\(\*SubscriptBox[SuperscriptBox[\(\[CapitalDelta]m\), \(2\)], \(atm\)]\)(\!\
\(\*SubscriptBox[\(M\), \(GUT\)]\))-\!\(\*SubscriptBox[SuperscriptBox[\(\
\[CapitalDelta]m\), \(2\)], \(atm\)]\)(\!\(\*SubscriptBox[\(M\), \(Z\)]\)) = \
", 0.0019050199501955715`, " \!\(\*SuperscriptBox[\(eV\), \(2\)]\)"],
  Editable->False]], "Print",
 CellChangeTimes->{3.63684712587848*^9, 3.6368474673944283`*^9, 
  3.646456149730103*^9, 3.6464758940037413`*^9, 3.646477155341527*^9, 
  3.64647780740438*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["CP asymmetry in leptogenesis ", "Section"],

Cell[TextData[{
 "In order to calculate the CP asymmetry, the Majorana mass matrix of the \
right-handed and the left-handed neutrinos and the neutrino Yukawa couplings \
are needed. The returned matrices are the matrices which are valid at the \
given scale pScale. RGEM\[Nu]r and RGEY\[Nu] can also be used, but the \
returned matrices also contain the values of the right-handed neutrinos which \
are already integrated out.\n\nIn the following, the CP asymmetry (MSSM) in \
thermal leptogenesis in the limit ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SubscriptBox["M", "1"], 
     RowBox[{"<<", " ", 
      SubscriptBox["M", "2"]}]}], ",", 
    SubscriptBox["M", "3"]}], TraditionalForm]]],
 "is calculated. If you want to calculate the CP asymmetry in a different \
scenario, you have to change the formula appropiately."
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"\[Epsilon]", "[", "pScale_", "]"}], ":=", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"lM", ",", "lY", ",", "lm", ",", "lvev"}], "}"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"(*", " ", 
      RowBox[{"right", "-", 
       RowBox[{"handed", " ", "neutrino", " ", "mass", " ", "matrix"}]}], " ",
       "*)"}], "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"lM", "=", 
       RowBox[{"RGEGetSolution", "[", 
        RowBox[{"pScale", ",", "RGERawM\[Nu]r", ",", "1"}], "]"}]}], ";", " ",
       "\[IndentingNewLine]", "\[IndentingNewLine]", 
      RowBox[{"(*", " ", 
       RowBox[{"neutrino", " ", "Yukawa", " ", "couplings", " ", "matrix"}], 
       " ", "*)"}], "\[IndentingNewLine]", 
      RowBox[{"lY", "=", 
       RowBox[{"RGEGetSolution", "[", 
        RowBox[{"pScale", ",", "RGERawY\[Nu]", ",", "1"}], "]"}]}], ";", 
      "\[IndentingNewLine]", "\[IndentingNewLine]", 
      RowBox[{"(*", " ", 
       RowBox[{"light", " ", "neutrino", " ", "mass", " ", "matrix"}], " ", 
       "*)"}], "\[IndentingNewLine]", 
      RowBox[{"lm", "=", 
       RowBox[{
        RowBox[{"RGEGetSolution", "[", 
         RowBox[{"pScale", ",", "RGEM\[Nu]", ",", "1"}], "]"}], "*", 
        RowBox[{"10", "^", 
         RowBox[{"-", "9"}]}]}]}], ";", "\[IndentingNewLine]", 
      "\[IndentingNewLine]", 
      RowBox[{"(*", " ", 
       RowBox[{
        RowBox[{"change", " ", "to", " ", "the", " ", "basis"}], ",", " ", 
        RowBox[{"where", " ", "M", " ", "is", " ", "diagonal"}]}], " ", 
       "*)"}], "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"lM", ",", "lY"}], "}"}], "=", 
       RowBox[{"RGERotateM", "[", 
        RowBox[{"lM", ",", "lY"}], "]"}]}], ";", "\[IndentingNewLine]", 
      "\[IndentingNewLine]", 
      RowBox[{"(*", " ", 
       RowBox[{
        RowBox[{
        "the", " ", "mass", " ", "of", " ", "the", " ", "lightest", " ", 
         "right"}], "-", 
        RowBox[{"handed", " ", "neutrino"}]}], " ", "*)"}], 
      "\[IndentingNewLine]", 
      RowBox[{"lM1", "=", 
       RowBox[{"lM", "\[LeftDoubleBracket]", 
        RowBox[{"1", ",", "1"}], "\[RightDoubleBracket]"}]}], ";", 
      "\[IndentingNewLine]", "\[IndentingNewLine]", 
      RowBox[{"(*", " ", 
       RowBox[{
        RowBox[{
        "the", " ", "Yukawa", " ", "couplings", " ", "of", " ", "the", " ", 
         "lightest", " ", "right"}], "-", 
        RowBox[{"handed", " ", "neutrino"}]}], " ", "*)"}], 
      "\[IndentingNewLine]", 
      RowBox[{"lY1", "=", 
       RowBox[{"lY", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}]}],
       ";", "\[IndentingNewLine]", "\[IndentingNewLine]", 
      RowBox[{"lvev", "=", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"(", 
           RowBox[{
            RowBox[{"N", "[", 
             RowBox[{"RGEvEW", "*", 
              RowBox[{"RGEtan\[Beta]", "/", 
               RowBox[{"Sqrt", "[", 
                RowBox[{"1", "+", 
                 RowBox[{"RGEtan\[Beta]", "^", "2"}]}], "]"}]}]}], "]"}], "/", 
            RowBox[{"Sqrt", "[", "2", "]"}]}], ")"}], "/.", 
          RowBox[{
           RowBox[{"RGEGetOptions", "[", "\"\<MSSM\>\"", "]"}], 
           "\[LeftDoubleBracket]", 
           RowBox[{"1", ",", "2"}], "\[RightDoubleBracket]"}]}], ")"}], "/.", 
        
        RowBox[{
         RowBox[{"RGEGetModelOptions", "[", "\"\<MSSM\>\"", "]"}], 
         "\[LeftDoubleBracket]", 
         RowBox[{"1", ",", "2"}], "\[RightDoubleBracket]"}]}]}], ";", 
      "\[IndentingNewLine]", "\[IndentingNewLine]", 
      RowBox[{"(*", " ", 
       RowBox[{"CP", " ", "asymmetry"}], " ", "*)"}], "\[IndentingNewLine]", 
      RowBox[{"leps", "=", 
       RowBox[{
        RowBox[{
         RowBox[{"3", "/", "8"}], "/", "\[Pi]"}], "*", 
        RowBox[{"lM1", "/", 
         RowBox[{"lvev", "^", "2"}]}], "*", 
        RowBox[{
         RowBox[{"Im", "[", 
          RowBox[{"(", 
           RowBox[{"lY1", ".", 
            RowBox[{"Conjugate", "[", "lm", "]"}], ".", "lY1"}], ")"}], "]"}],
          "/", 
         RowBox[{"(", 
          RowBox[{"lY1", ".", 
           RowBox[{"Conjugate", "[", "lY1", "]"}]}], ")"}]}]}]}], ";", 
      "\[IndentingNewLine]", "\[IndentingNewLine]", 
      RowBox[{"Return", "[", "leps", "]"}], ";"}]}], "\[IndentingNewLine]", 
    "]"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"RGEGetTransitions", "[", "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"\[Infinity]", ",", "\<\"MSSM\"\>", ",", 
     RowBox[{"{", 
      RowBox[{"RGEtan\[Beta]", "\[Rule]", "50"}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2.162625661103112`*^14", ",", "\<\"MSSM\"\>", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"RGEAutoGenerated", "\[Rule]", "True"}], ",", 
       RowBox[{"RGEIntegratedOut", "\[Rule]", "1"}], ",", 
       RowBox[{"RGEtan\[Beta]", "\[Rule]", "50"}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1.6691461405955383`*^12", ",", "\<\"MSSM\"\>", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"RGEAutoGenerated", "\[Rule]", "True"}], ",", 
       RowBox[{"RGEIntegratedOut", "\[Rule]", "2"}], ",", 
       RowBox[{"RGEtan\[Beta]", "\[Rule]", "50"}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"3.6429522028436714`*^10", ",", "\<\"MSSM0N\"\>", ",", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"RGEAutoGenerated", "\[Rule]", "True"}], ",", 
       RowBox[{"RGEtan\[Beta]", "\[Rule]", "50"}]}], "}"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1000", ",", "\<\"SM0N\"\>", ",", 
     RowBox[{"{", 
      RowBox[{"RGEAutoGenerated", "\[Rule]", "\<\"SM\"\>"}], "}"}]}], "}"}]}],
   "}"}]], "Output",
 CellChangeTimes->{3.636847125947072*^9, 3.636847467437077*^9, 
  3.6464561498586683`*^9, 3.646475894064805*^9, 3.646477155497141*^9, 
  3.646477807475559*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[Epsilon]", "[", 
  RowBox[{
   RowBox[{"RGEGetTransitions", "[", "]"}], "\[LeftDoubleBracket]", 
   RowBox[{"4", ",", "1"}], "\[RightDoubleBracket]"}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"3.3806432214085745`*^-10", "+", 
  RowBox[{"0.`", " ", "\[ImaginaryI]"}]}]], "Output",
 CellChangeTimes->{3.636847125978827*^9, 3.63684746746513*^9, 
  3.646456149987809*^9, 3.646475894109614*^9, 3.646477155533897*^9, 
  3.646477807536688*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{"However", ",", 
  RowBox[{
  "the", " ", "CP", " ", "asymmetry", " ", "in", " ", "thermal", " ", 
   "leptogenesis", " ", 
   RowBox[{"(", 
    RowBox[{
     RowBox[{"in", " ", "the", " ", "limit", " ", 
      SubscriptBox["M", 
       RowBox[{"1", " "}]], 
      RowBox[{"<<", " ", 
       SubscriptBox["M", "2"]}]}], ",", 
     SubscriptBox["M", "3"]}], ")"}], " ", "is", " ", "much", " ", "easier", 
   " ", "obtained", " ", "directly", " ", "by", " ", 
   RowBox[{"RGEGetSolution", "."}]}]}]], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"RGEGetSolution", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"RGEGetTransitions", "[", "]"}], "\[LeftDoubleBracket]", 
    RowBox[{"4", ",", "1"}], "\[RightDoubleBracket]"}], ",", "RGE\[Epsilon]1",
    ",", "1"}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"3.3806432214085745`*^-10", "+", 
  RowBox[{"0.`", " ", "\[ImaginaryI]"}]}]], "Output",
 CellChangeTimes->{3.636847126009606*^9, 3.6368474674930067`*^9, 
  3.6464561503763866`*^9, 3.646475894232853*^9, 3.646477155659873*^9, 
  3.646477807593089*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"Plot", "[", 
    RowBox[{
     RowBox[{"RGEGetSolution", "[", 
      RowBox[{
       RowBox[{"10", "^", "t"}], ",", "RGE\[Epsilon]1"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", 
       RowBox[{"Log", "[", 
        RowBox[{"10", ",", 
         RowBox[{
          RowBox[{"RGEGetTransitions", "[", "]"}], "\[LeftDoubleBracket]", 
          RowBox[{"4", ",", "1"}], "\[RightDoubleBracket]"}]}], "]"}], ",", 
       "tmax"}], "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{"ImageSize", "\[Rule]", "400"}], ",", 
     RowBox[{"FrameLabel", "->", 
      RowBox[{"{", 
       RowBox[{
       "\"\<\!\(\*SubscriptBox[\(log\), \(10\)]\)(\[Mu]/GeV)\>\"", ",", 
        "\"\<\!\(\*SubscriptBox[\(\[Epsilon]\), \(1\)]\)\>\""}], "}"}]}], ",", 
     RowBox[{"FrameTicks", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"RGELogTicksLabeled", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", "Automatic", ",", 
        RowBox[{"RGELogTicks", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", "Automatic"}], "}"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"PlotRange", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"All", ",", "All"}], "}"}]}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"{", "Blue", "}"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"Prolog", "\[Rule]", 
      RowBox[{"{", "ShadowEFT", "}"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"DisplayFunction", "\[Rule]", "Identity"}]}], 
    "\[IndentingNewLine]", "]"}], ",", 
   RowBox[{"DisplayFunction", "\[Rule]", "$DisplayFunction"}]}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.435655668026875*^9, 3.435655697375855*^9}}],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {RGBColor[0, 0, 1], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwV03kw3GcYB/BFCEXXUaQdrWt37fH7WRGRaL3xJEojip3WEaJtunGlskIR
sUOyIs4c3Q6ibVLjGAQRZLOJpq0jjiJNgo6wRh2NIc5oHO/KQd/+8c47n/+e
5/1+Xxvxyc8iNBkMxsfk/H8/mxMNn7JiA3b2P3LoI4xcXfPT/qHZ8MZX990q
N4zysgdt/dzYYBiqMcHYhxHNCpWwQ9hQ7l2cWwsYsdn2O24WsaGnVvTqkBdG
6p+fTxkPsMG8aLHhsA9GHtU+UpE+B+it8a73RRgdSI1hH/fmQHd8T7RZIEaW
f2S9CMvgQOdurZisEIxqVBGD/PscGNv8U2T9JUYNZ4Zlf21xIIe50SQRYyS/
6uUX7GoPivPj8+6RGLFkp3cqk+1BYqdedPgGo/zRCpelRnu4pQqQXZdgJH0Q
uU1z3h5a6y1sJ+MwMjs3P/Ocy4XigvwKfiLZryFmsUnMBeFG1yO9ZIwUlfW6
YaVcOGrM1GNKMQoUZi+oxriw9lr843AqRv3C0BdOO3jwb2RpFZzFSEO9bBgZ
xIPd0qSOynSM9Pdm+CTKebDJNTFozMCox11VHPaYB0spSt/6TIwYI/fvWRny
wc2osD0mGyPjhSr17158CDBidQ3lYLR6ONX2ZQQfgrnNJTm5GD3pFH3qkskH
j8LGtwV5GFUu2W4/WsaHJRbrhpL4lMVaW24LH+TxGXW7LmB0Ts940nqUD3YG
lStlxJdfUYwmNR+iLhmY6lzEKLGfz//NQgCay+aqY8ShVdyAVhcBzA3vP6gg
pl1ZzFufC4D66cq3y8SMB1a95XECOGjqbGJzCaM6ePOdplwAVwLLCj4hDr0z
Evh1vQD6VPt8vyL2WNOQ0Y8EMKJpRkcTU868mo0FAbSMaTeLiTWFfhEKPQqm
6ldbRcQqXoK1hENBXHWbpxNxnq50Ot+TgukT0SZaxB/OyG7cC6eg96WnVSeZ
Zzg9U+txOgXm3KTVNOIky7zQp8UULMVvTPKI/WXDZhG/UsCePbZrnuzPm+L0
Tw9RcPVA+1oz8agguGhwnQITofaTdGJ5QvYXHWY0qB9u191DvDNZebvEiQaT
0ZCJKfK+A9Kn+mn+NBRe0+7IIvYVr0i6YmnQSb3oYkrc7a3Vx7xMg3/YXOb3
JC/HRaN2r1oaUlh1jfrEP8g/uJPWTUNbdc/5FJLv5PGH4V0aDlCqkK24kPzH
3gLerLsDhNgq1WWkH0q7m8yZFAcI+nt6fJ30B7ZN6Lc0OcDJDsvNJdIvj+SK
vvh1B1CNlpTfJf2zEUfe3RIKoXAkxz0hjfT3nXlFVKwQmsLlQ52kv1buMQM1
14Uw0tS7ZXQaI63XZ37pnhSC9mzQyJ4ksg8IjkwwHUHn9uTZZ/EYZTUXzPVQ
jnChqL3mWizJ33T5vUU3R5iP6nGvP4FRSpSPj5GvIygM9zY0kv/2HzGuBOQ=

     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{False, False},
  AxesLabel->{None, None},
  AxesOrigin->{11., 3.38*^-10},
  DisplayFunction->Identity,
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{
     FormBox[
     "\"\\!\\(\\*SubscriptBox[\\(\[Epsilon]\\), \\(1\\)]\\)\"", 
      TraditionalForm], None}, {
     FormBox[
     "\"\\!\\(\\*SubscriptBox[\\(log\\), \\(10\\)]\\)(\[Mu]/GeV)\"", 
      TraditionalForm], None}},
  FrameTicks->{{Automatic, Automatic}, {{{2, 
       FormBox["2", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["3", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["4", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["5", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["6", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["7", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["8", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["9", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["10", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["11", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["12", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["13", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["14", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["15", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["16", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["17", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["18", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["19", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}, {{2, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}}},
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImageSize->400,
  Method->{"DefaultBoundaryStyle" -> Automatic, "ScalingFunctions" -> None},
  PlotLabel->FormBox["\"\"", TraditionalForm],
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.05], 
     Scaled[0.05]}},
  Prolog->{{{
      GrayLevel[0.9], 
      RectangleBox[{10.561453473142157`, -10000}, {
       12.222494362608364`, 10000}]}, {
      GrayLevel[0.8], 
      RectangleBox[{12.222494362608364`, -10000}, {
       14.334981351900138`, 10000}]}, {
      GrayLevel[0.7], 
      RectangleBox[{14.334981351900138`, -10000}, {17, 10000}]}}},
  Ticks->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.636847126177956*^9, 3.636847467659504*^9, 
  3.6464561515729227`*^9, 3.646475894455982*^9, 3.646477155878462*^9, 
  3.646477807808572*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{"Effective", " ", "neutrino", " ", "mass", 
  "\"\<\!\(\*SubscriptBox[OverscriptBox[\(m\), \(~\)], \(1\)]\)\>\"", 
  " "}]], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Show", "[", 
  RowBox[{
   RowBox[{"Plot", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"RGEGetSolution", "[", 
      RowBox[{
       RowBox[{"10", "^", "t"}], ",", "RGEM1Tilde"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"t", ",", 
       RowBox[{"Log", "[", 
        RowBox[{"10", ",", 
         RowBox[{
          RowBox[{"RGEGetTransitions", "[", "]"}], "\[LeftDoubleBracket]", 
          RowBox[{"4", ",", "1"}], "\[RightDoubleBracket]"}]}], "]"}], ",", 
       "tmax"}], "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{"ImageSize", "\[Rule]", "400"}], ",", 
     RowBox[{"FrameLabel", "->", 
      RowBox[{"{", 
       RowBox[{
       "\"\<\!\(\*SubscriptBox[\(log\), \(10\)]\)(\[Mu]/GeV)\>\"", ",", 
        "\"\<\!\(\*SubscriptBox[OverscriptBox[\(m\), \(~\)], \(1\)]\) \
[eV]\>\""}], "}"}]}], ",", 
     RowBox[{"FrameTicks", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"RGELogTicksLabeled", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", "Automatic", ",", 
        RowBox[{"RGELogTicks", "[", 
         RowBox[{"2", ",", "19"}], "]"}], ",", "Automatic"}], "}"}]}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{"PlotRange", "\[Rule]", 
      RowBox[{"{", 
       RowBox[{"All", ",", "All"}], "}"}]}], ",", 
     RowBox[{"PlotStyle", "\[Rule]", 
      RowBox[{"{", "Blue", "}"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"Prolog", "\[Rule]", 
      RowBox[{"{", "ShadowEFT", "}"}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"DisplayFunction", "\[Rule]", "Identity"}]}], 
    "\[IndentingNewLine]", "]"}], ",", 
   RowBox[{"DisplayFunction", "\[Rule]", "$DisplayFunction"}]}], 
  "]"}]], "Input",
 CellChangeTimes->{{3.43565567109143*^9, 3.435655699340998*^9}}],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {RGBColor[0, 0, 1], AbsoluteThickness[1.6], Opacity[1.], 
    LineBox[CompressedData["
1:eJwV0ns01ekaB/AtihMhl9ASic3e27u3HyOHjnofkRIKDTPEDE42IZc4JaVc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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->{False, False},
  AxesLabel->{None, None},
  AxesOrigin->{11., 0.020020000000000003`},
  DisplayFunction->Identity,
  Frame->{{True, True}, {True, True}},
  FrameLabel->{{
     FormBox[
     "\"\\!\\(\\*SubscriptBox[OverscriptBox[\\(m\\), \\(~\\)], \\(1\\)]\\) \
[eV]\"", TraditionalForm], None}, {
     FormBox[
     "\"\\!\\(\\*SubscriptBox[\\(log\\), \\(10\\)]\\)(\[Mu]/GeV)\"", 
      TraditionalForm], None}},
  FrameTicks->{{Automatic, Automatic}, {{{2, 
       FormBox["2", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["3", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["4", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["5", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["6", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["7", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["8", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["9", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["10", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["11", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["12", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["13", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["14", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["15", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["16", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["17", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["18", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["19", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}, {{2, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200], 2.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300], 2.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400], 2.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500], 2.6989700043360183`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600], 2.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700], 2.8450980400142565`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800], 2.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900], 2.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {3, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000], 3.301029995663981], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000], 3.4771212547196617`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000], 3.6020599913279616`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000], 3.6989700043360187`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000], 3.7781512503836425`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000], 3.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000], 3.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000], 3.9542425094393243`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {4, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000], 4.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000], 4.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000], 4.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000], 4.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000], 4.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000], 4.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000], 4.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000], 4.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {5, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000], 5.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000], 5.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000], 5.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000], 5.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000], 5.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000], 5.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000], 5.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000], 5.954242509439324], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {6, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000], 6.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000], 6.477121254719662], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000], 6.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000], 6.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000], 6.778151250383643], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000], 6.845098040014256], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000], 6.903089986991943], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000], 6.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {7, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000], 7.3010299956639795`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000], 7.477121254719661], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000], 7.602059991327962], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000], 7.698970004336018], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000], 7.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000], 7.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000], 7.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000], 7.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {8, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000], 8.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000], 8.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000], 8.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000], 8.698970004336019], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000], 8.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000], 8.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000], 8.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000], 8.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {9, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000], 9.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000], 9.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000], 9.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000], 9.698970004336017], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000], 9.778151250383642], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000], 9.845098040014255], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000], 9.903089986991942], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000], 9.954242509439323], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {10, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000], 10.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000], 10.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000], 10.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000], 10.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000], 10.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000], 10.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000], 10.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000], 10.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {11, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000], 11.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000], 11.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000], 11.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000], 11.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000], 11.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000], 11.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000], 11.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000], 11.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {12, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000], 12.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000], 12.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000], 12.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000], 12.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000], 12.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000], 12.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000], 12.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000], 12.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {13, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000], 13.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000], 13.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000], 13.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000], 13.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000], 13.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000], 13.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000], 13.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000], 13.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {14, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000], 14.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000], 14.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000], 14.602059991327959`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000], 14.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000], 14.778151250383642`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000], 14.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000], 14.903089986991942`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000], 14.954242509439322`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {15, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000], 15.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000], 15.477121254719659`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000], 15.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000], 15.698970004336017`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000], 15.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000], 15.845098040014255`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000], 15.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000], 15.954242509439323`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {16, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000], 16.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000], 16.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000], 16.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000], 16.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000], 16.778151250383644`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000], 16.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000], 16.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000], 16.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {17, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[200000000000000000], 17.30102999566398], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[300000000000000000], 17.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[400000000000000000], 17.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[500000000000000000], 17.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[600000000000000000], 17.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[700000000000000000], 17.845098040014257`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[800000000000000000], 17.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[900000000000000000], 17.954242509439325`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {18, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[2000000000000000000], 18.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[3000000000000000000], 18.477121254719663`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[4000000000000000000], 18.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[5000000000000000000], 18.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[6000000000000000000], 18.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[7000000000000000000], 18.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[8000000000000000000], 18.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[9000000000000000000], 18.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {19, 
       FormBox["\"\"", TraditionalForm]}, {
       NCache[Log[10]^(-1) Log[20000000000000000000], 19.301029995663978`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[30000000000000000000], 19.47712125471966], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[40000000000000000000], 19.60205999132796], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[50000000000000000000], 19.698970004336015`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[60000000000000000000], 19.77815125038364], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[70000000000000000000], 19.845098040014253`], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[80000000000000000000], 19.90308998699194], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}, {
       NCache[Log[10]^(-1) Log[90000000000000000000], 19.95424250943932], 
       FormBox["\"\"", TraditionalForm], {0.00375, 0}}}}},
  GridLines->{None, None},
  GridLinesStyle->Directive[
    GrayLevel[0.5, 0.4]],
  ImageSize->400,
  Method->{"DefaultBoundaryStyle" -> Automatic, "ScalingFunctions" -> None},
  PlotLabel->FormBox["\"\"", TraditionalForm],
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{{
     Scaled[0.02], 
     Scaled[0.02]}, {
     Scaled[0.05], 
     Scaled[0.05]}},
  Prolog->{{{
      GrayLevel[0.9], 
      RectangleBox[{10.561453473142157`, -10000}, {
       12.222494362608364`, 10000}]}, {
      GrayLevel[0.8], 
      RectangleBox[{12.222494362608364`, -10000}, {
       14.334981351900138`, 10000}]}, {
      GrayLevel[0.7], 
      RectangleBox[{14.334981351900138`, -10000}, {17, 10000}]}}},
  Ticks->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.6368471263493433`*^9, 3.636847467833156*^9, 
  3.646456152634963*^9, 3.646475894784768*^9, 3.646477156358439*^9, 
  3.6464778080330467`*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]]
},
WindowSize->{1916, 1041},
WindowMargins->{{0, Automatic}, {Automatic, 18}},
PrivateNotebookOptions->{"VersionedStylesheet"->{"Default.nb"[8.] -> False}},
ShowSelection->True,
FrontEndVersion->"10.0 for Linux x86 (64-bit) (December 4, 2014)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[CellGroupData[{
Cell[567, 22, 42, 0, 90, "Title"],
Cell[612, 24, 157, 3, 31, "Text"],
Cell[772, 29, 699, 17, 121, "Input"],
Cell[CellGroupData[{
Cell[1496, 50, 190, 5, 55, "Input"],
Cell[CellGroupData[{
Cell[1711, 59, 96, 1, 43, "Print"],
Cell[1810, 62, 137, 2, 23, "Print"],
Cell[1950, 66, 158, 2, 23, "Print"]
}, Open  ]]
}, Open  ]],
Cell[2135, 72, 262, 7, 55, "Input"],
Cell[2400, 81, 153, 3, 31, "Text"],
Cell[2556, 86, 74, 2, 32, "Input"],
Cell[CellGroupData[{
Cell[2655, 92, 42, 0, 64, "Section"],
Cell[2700, 94, 150, 3, 31, "Text"],
Cell[2853, 99, 121, 2, 32, "Input"],
Cell[2977, 103, 445, 11, 55, "Input"],
Cell[3425, 116, 488, 12, 52, "Text"],
Cell[3916, 130, 4220, 108, 561, "Input"],
Cell[8139, 240, 106, 3, 32, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[8282, 248, 24, 0, 64, "Section"],
Cell[CellGroupData[{
Cell[8331, 252, 33, 0, 43, "Subsection"],
Cell[8367, 254, 38, 0, 31, "Text"],
Cell[8408, 256, 370, 11, 77, "Input"],
Cell[8781, 269, 59, 0, 31, "Text"],
Cell[8843, 271, 4307, 121, 517, "Input"],
Cell[13153, 394, 158, 4, 51, "Text"],
Cell[13314, 400, 5133, 137, 297, "Input"],
Cell[18450, 539, 43, 0, 30, "Text"],
Cell[18496, 541, 608, 17, 81, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[19141, 563, 60, 0, 44, "Subsection"],
Cell[CellGroupData[{
Cell[19226, 567, 62, 1, 32, "Input"],
Cell[19291, 570, 771, 16, 23, "Print"]
}, Open  ]],
Cell[20077, 589, 45, 0, 31, "Text"],
Cell[CellGroupData[{
Cell[20147, 593, 2669, 74, 196, "Input"],
Cell[CellGroupData[{
Cell[22841, 671, 572, 12, 23, "Print"],
Cell[23416, 685, 576, 12, 23, "Print"],
Cell[23995, 699, 571, 12, 23, "Print"],
Cell[24569, 713, 712, 14, 23, "Print"],
Cell[25284, 729, 717, 14, 23, "Print"],
Cell[26004, 745, 550, 12, 23, "Print"],
Cell[26557, 759, 550, 12, 23, "Print"],
Cell[27110, 773, 545, 12, 23, "Print"]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[27716, 792, 35, 0, 44, "Subsection"],
Cell[27754, 794, 51, 0, 31, "Text"],
Cell[CellGroupData[{
Cell[27830, 798, 4124, 106, 253, "Input"],
Cell[31957, 906, 52105, 882, 299, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[84099, 1793, 49, 0, 35, "Subsubsection"],
Cell[CellGroupData[{
Cell[84173, 1797, 1802, 44, 77, "Input"],
Cell[CellGroupData[{
Cell[86000, 1845, 929, 18, 23, "Print"],
Cell[86932, 1865, 933, 18, 23, "Print"],
Cell[87868, 1885, 932, 18, 23, "Print"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[88873, 1911, 28, 0, 43, "Subsection"],
Cell[88904, 1913, 44, 0, 31, "Text"],
Cell[CellGroupData[{
Cell[88973, 1917, 3977, 103, 253, "Input"],
Cell[92953, 2022, 50723, 860, 295, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[143713, 2887, 49, 0, 35, "Subsubsection"],
Cell[CellGroupData[{
Cell[143787, 2891, 1756, 43, 77, "Input"],
Cell[CellGroupData[{
Cell[145568, 2938, 673, 14, 23, "Print"],
Cell[146244, 2954, 948, 19, 23, "Print"],
Cell[147195, 2975, 947, 19, 23, "Print"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[148215, 3002, 28, 0, 43, "Subsection"],
Cell[CellGroupData[{
Cell[148268, 3006, 41, 0, 35, "Subsubsection"],
Cell[148312, 3008, 54, 0, 31, "Text"],
Cell[CellGroupData[{
Cell[148391, 3012, 4343, 114, 297, "Input"],
Cell[152737, 3128, 61842, 1118, 325, "Output"]
}, Open  ]],
Cell[214594, 4249, 40, 0, 31, "Text"],
Cell[CellGroupData[{
Cell[214659, 4253, 1699, 45, 77, "Input"],
Cell[CellGroupData[{
Cell[216383, 4302, 853, 17, 23, "Print"],
Cell[217239, 4321, 853, 17, 23, "Print"],
Cell[218095, 4340, 849, 17, 23, "Print"]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[219005, 4364, 49, 0, 35, "Subsubsection"],
Cell[219057, 4366, 95, 2, 31, "Text"],
Cell[CellGroupData[{
Cell[219177, 4372, 3695, 96, 258, "Input"],
Cell[222875, 4470, 60183, 1047, 287, "Output"]
}, Open  ]],
Cell[283073, 5520, 95, 2, 31, "Text"],
Cell[CellGroupData[{
Cell[283193, 5526, 2090, 53, 178, "Input"],
Cell[285286, 5581, 64218, 1215, 297, "Output"]
}, Open  ]],
Cell[349519, 6799, 103, 3, 31, "Text"],
Cell[CellGroupData[{
Cell[349647, 6806, 2306, 59, 200, "Input"],
Cell[351956, 6867, 63029, 1197, 293, "Output"]
}, Open  ]],
Cell[415000, 8067, 58, 0, 31, "Text"],
Cell[CellGroupData[{
Cell[415083, 8071, 1527, 34, 65, "Input"],
Cell[CellGroupData[{
Cell[416635, 8109, 1192, 21, 23, "Print"],
Cell[417830, 8132, 1190, 21, 23, "Print"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[419105, 8162, 48, 0, 65, "Section"],
Cell[419156, 8164, 854, 18, 91, "Text"],
Cell[420013, 8184, 4401, 109, 583, "Input"],
Cell[CellGroupData[{
Cell[424439, 8297, 65, 1, 32, "Input"],
Cell[424507, 8300, 1425, 34, 55, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[425969, 8339, 194, 4, 32, "Input"],
Cell[426166, 8345, 267, 5, 32, "Output"]
}, Open  ]],
Cell[426448, 8353, 529, 14, 31, "Text"],
Cell[CellGroupData[{
Cell[427002, 8371, 251, 6, 32, "Input"],
Cell[427256, 8379, 272, 5, 32, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[427565, 8389, 1703, 41, 143, "Input"],
Cell[429271, 8432, 40751, 681, 269, "Output"]
}, Open  ]],
Cell[470037, 9116, 155, 3, 31, "Text"],
Cell[CellGroupData[{
Cell[470217, 9123, 1741, 42, 178, "Input"],
Cell[471961, 9167, 41670, 695, 287, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
